# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
import json
import logging
import os
import time
from concurrent.futures import Executor
from functools import partial
from os.path import dirname

from conda.core.index import get_index
from conda_index.index import update_index as _update_index

from . import conda_interface, utils
from .conda_interface import CondaHTTPError, context, url_path
from .deprecations import deprecated
from .utils import (
    CONDA_PACKAGE_EXTENSION_V1,
    CONDA_PACKAGE_EXTENSION_V2,
    JSONDecodeError,
    get_logger,
    on_win,
)

log = get_logger(__name__)


@deprecated("24.3", "24.5")
class DummyExecutor(Executor):
    def map(self, func, *iterables):
        for iterable in iterables:
            for thing in iterable:
                yield func(thing)


local_index_timestamp = 0
cached_index = None
local_subdir = ""
local_output_folder = ""
cached_channels = []
_channel_data = {}
deprecated.constant("24.1", "24.5", "channel_data", _channel_data)


# TODO: support for libarchive seems to have broken ability to use multiple threads here.
#    The new conda format is so much faster that it more than makes up for it.  However, it
#    would be nice to fix this at some point.
_MAX_THREADS_DEFAULT = os.cpu_count() or 1
if on_win:  # see https://github.com/python/cpython/commit/8ea0fd85bc67438f679491fae29dfe0a3961900a
    _MAX_THREADS_DEFAULT = min(48, _MAX_THREADS_DEFAULT)
deprecated.constant("24.3", "24.5", "MAX_THREADS_DEFAULT", _MAX_THREADS_DEFAULT)
deprecated.constant("24.3", "24.5", "LOCK_TIMEOUT_SECS", 3 * 3600)
deprecated.constant("24.3", "24.5", "LOCKFILE_NAME", ".lock")

# TODO: this is to make sure that the index doesn't leak tokens.  It breaks use of private channels, though.
# os.environ['CONDA_ADD_ANACONDA_TOKEN'] = "false"


def get_build_index(
    subdir,
    bldpkgs_dir,
    output_folder=None,
    clear_cache=False,
    omit_defaults=False,
    channel_urls=None,
    debug=False,
    verbose=True,
    locking=None,
    timeout=None,
):
    """
    Used during package builds to create/get a channel including any local or
    newly built packages. This function both updates and gets index data.
    """
    global local_index_timestamp
    global local_subdir
    global local_output_folder
    global cached_index
    global cached_channels
    global _channel_data
    mtime = 0

    channel_urls = list(utils.ensure_list(channel_urls))

    if not output_folder:
        output_folder = dirname(bldpkgs_dir)

    # check file modification time - this is the age of our local index.
    index_file = os.path.join(output_folder, subdir, "repodata.json")
    if os.path.isfile(index_file):
        mtime = os.path.getmtime(index_file)

    if (
        clear_cache
        or not os.path.isfile(index_file)
        or local_subdir != subdir
        or local_output_folder != output_folder
        or mtime > local_index_timestamp
        or cached_channels != channel_urls
    ):
        # priority: (local as either croot or output_folder IF NOT EXPLICITLY IN CHANNEL ARGS),
        #     then channels passed as args (if local in this, it remains in same order),
        #     then channels from condarc.
        urls = list(channel_urls)

        loggers = utils.LoggingContext.default_loggers + [__name__]
        if debug:
            log_context = partial(utils.LoggingContext, logging.DEBUG, loggers=loggers)
        elif verbose:
            log_context = partial(utils.LoggingContext, logging.WARN, loggers=loggers)
        else:
            log_context = partial(
                utils.LoggingContext, logging.CRITICAL + 1, loggers=loggers
            )
        with log_context():
            # this is where we add the "local" channel.  It's a little smarter than conda, because
            #     conda does not know about our output_folder when it is not the default setting.
            if os.path.isdir(output_folder):
                local_path = url_path(output_folder)
                # replace local with the appropriate real channel.  Order is maintained.
                urls = [url if url != "local" else local_path for url in urls]
                if local_path not in urls:
                    urls.insert(0, local_path)
            _ensure_valid_channel(output_folder, subdir)
            _delegated_update_index(output_folder, verbose=debug)

            # replace noarch with native subdir - this ends up building an index with both the
            #      native content and the noarch content.

            if subdir == "noarch":
                subdir = conda_interface.subdir
            try:
                # get_index() is like conda reading the index, not conda_index
                # creating a new index.
                cached_index = get_index(
                    channel_urls=urls,
                    prepend=not omit_defaults,
                    use_local=False,
                    use_cache=context.offline,
                    platform=subdir,
                )
            # HACK: defaults does not have the many subfolders we support.  Omit it and
            #          try again.
            except CondaHTTPError:
                if "defaults" in urls:
                    urls.remove("defaults")
                cached_index = get_index(
                    channel_urls=urls,
                    prepend=omit_defaults,
                    use_local=False,
                    use_cache=context.offline,
                    platform=subdir,
                )

            expanded_channels = {rec.channel for rec in cached_index}

            superchannel = {}
            # we need channeldata.json too, as it is a more reliable source of run_exports data
            for channel in expanded_channels:
                if channel.scheme == "file":
                    location = channel.location
                    if utils.on_win:
                        location = location.lstrip("/")
                    elif not os.path.isabs(channel.location) and os.path.exists(
                        os.path.join(os.path.sep, channel.location)
                    ):
                        location = os.path.join(os.path.sep, channel.location)
                    channeldata_file = os.path.join(
                        location, channel.name, "channeldata.json"
                    )
                    retry = 0
                    max_retries = 1
                    if os.path.isfile(channeldata_file):
                        while retry < max_retries:
                            try:
                                with open(channeldata_file, "r+") as f:
                                    _channel_data[channel.name] = json.load(f)
                                break
                            except (OSError, JSONDecodeError):
                                time.sleep(0.2)
                                retry += 1
                else:
                    # download channeldata.json for url
                    if not context.offline:
                        try:
                            _channel_data[channel.name] = utils.download_channeldata(
                                channel.base_url + "/channeldata.json"
                            )
                        except CondaHTTPError:
                            continue
                # collapse defaults metachannel back into one superchannel, merging channeldata
                if channel.base_url in context.default_channels and _channel_data.get(
                    channel.name
                ):
                    packages = superchannel.get("packages", {})
                    packages.update(_channel_data[channel.name])
                    superchannel["packages"] = packages
            _channel_data["defaults"] = superchannel
        local_index_timestamp = os.path.getmtime(index_file)
        local_subdir = subdir
        local_output_folder = output_folder
        cached_channels = channel_urls
    return cached_index, local_index_timestamp, _channel_data


def _ensure_valid_channel(local_folder, subdir):
    for folder in {subdir, "noarch"}:
        path = os.path.join(local_folder, folder)
        if not os.path.isdir(path):
            os.makedirs(path)


def _delegated_update_index(
    dir_path,
    check_md5=False,
    channel_name=None,
    patch_generator=None,
    threads=1,
    verbose=False,
    progress=False,
    subdirs=None,
    warn=True,
    current_index_versions=None,
    debug=False,
):
    """
    update_index as called by conda-build, delegating to standalone conda-index.
    Needed to allow update_index calls on single subdir.
    """
    # conda-build calls update_index on a single subdir internally, but
    # conda-index expects to index every subdir under dir_path
    parent_path, dirname = os.path.split(dir_path)
    if dirname in utils.DEFAULT_SUBDIRS:
        dir_path = parent_path
        subdirs = [dirname]

    log_level = logging.DEBUG if debug else logging.INFO if verbose else logging.WARNING
    with utils.LoggingContext(log_level):
        return _update_index(
            dir_path,
            check_md5=check_md5,
            channel_name=channel_name,
            patch_generator=patch_generator,
            threads=threads,
            verbose=verbose,
            progress=progress,
            subdirs=subdirs,
            warn=warn,
            current_index_versions=current_index_versions,
            debug=debug,
        )


@deprecated(
    "24.1.0", "24.5.0", addendum="Use `conda_index._apply_instructions` instead."
)
def _apply_instructions(subdir, repodata, instructions):
    repodata.setdefault("removed", [])
    utils.merge_or_update_dict(
        repodata.get("packages", {}),
        instructions.get("packages", {}),
        merge=False,
        add_missing_keys=False,
    )
    # we could have totally separate instructions for .conda than .tar.bz2, but it's easier if we assume
    #    that a similarly-named .tar.bz2 file is the same content as .conda, and shares fixes
    new_pkg_fixes = {
        k.replace(CONDA_PACKAGE_EXTENSION_V1, CONDA_PACKAGE_EXTENSION_V2): v
        for k, v in instructions.get("packages", {}).items()
    }

    utils.merge_or_update_dict(
        repodata.get("packages.conda", {}),
        new_pkg_fixes,
        merge=False,
        add_missing_keys=False,
    )
    utils.merge_or_update_dict(
        repodata.get("packages.conda", {}),
        instructions.get("packages.conda", {}),
        merge=False,
        add_missing_keys=False,
    )

    for fn in instructions.get("revoke", ()):
        for key in ("packages", "packages.conda"):
            if fn.endswith(CONDA_PACKAGE_EXTENSION_V1) and key == "packages.conda":
                fn = fn.replace(CONDA_PACKAGE_EXTENSION_V1, CONDA_PACKAGE_EXTENSION_V2)
            if fn in repodata[key]:
                repodata[key][fn]["revoked"] = True
                repodata[key][fn]["depends"].append("package_has_been_revoked")

    for fn in instructions.get("remove", ()):
        for key in ("packages", "packages.conda"):
            if fn.endswith(CONDA_PACKAGE_EXTENSION_V1) and key == "packages.conda":
                fn = fn.replace(CONDA_PACKAGE_EXTENSION_V1, CONDA_PACKAGE_EXTENSION_V2)
            popped = repodata[key].pop(fn, None)
            if popped:
                repodata["removed"].append(fn)
    repodata["removed"].sort()

    return repodata



# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import configparser  # noqa: F401
import os
from functools import partial
from importlib import import_module  # noqa: F401

from conda import __version__ as CONDA_VERSION  # noqa: F401
from conda.base.context import context, determine_target_prefix, reset_context
from conda.base.context import non_x86_machines as non_x86_linux_machines  # noqa: F401
from conda.core.package_cache import ProgressiveFetchExtract  # noqa: F401
from conda.exceptions import (  # noqa: F401
    CondaError,
    CondaHTTPError,
    LinkError,
    LockError,
    NoPackagesFoundError,
    PaddingError,
    UnsatisfiableError,
)
from conda.exports import (  # noqa: F401
    ArgumentParser,
    Channel,
    Completer,
    CondaSession,
    EntityEncoder,
    FileMode,
    InstalledPackages,
    MatchSpec,
    NoPackagesFound,
    PackageRecord,
    PathType,
    Resolve,
    StringIO,
    TemporaryDirectory,
    TmpDownload,
    Unsatisfiable,
    VersionOrder,
    _toposort,
    add_parser_channels,
    add_parser_prefix,
    download,
    human_bytes,
    input,
    lchmod,
    normalized_version,
    prefix_placeholder,
    rm_rf,
    spec_from_line,
    specs_from_args,
    specs_from_url,
    symlink_conda,
    unix_path_to_win,
    untracked,
    url_path,
    walk_prefix,
    win_path_to_unix,
)
from conda.exports import get_index as _get_index
from conda.gateways.disk.read import compute_sum
from conda.models.channel import get_conda_build_local_url  # noqa: F401

from .deprecations import deprecated

deprecated.constant("24.1.0", "24.5.0", "get_index", _get_index)
# TODO: Go to references of all properties below and import them from `context` instead
binstar_upload = context.binstar_upload
default_python = context.default_python
envs_dirs = context.envs_dirs
pkgs_dirs = list(context.pkgs_dirs)
cc_platform = context.platform
root_dir = context.root_dir
root_writable = context.root_writable
subdir = context.subdir
create_default_packages = context.create_default_packages

get_rc_urls = lambda: list(context.channels)
get_prefix = partial(determine_target_prefix, context)
cc_conda_build = context.conda_build if hasattr(context, "conda_build") else {}

get_conda_channel = Channel.from_value

# Disallow softlinks. This avoids a lot of dumb issues, at the potential cost of disk space.
os.environ["CONDA_ALLOW_SOFTLINKS"] = "false"
reset_context()

# When deactivating envs (e.g. switching from root to build/test) this env var is used,
# except the PR that removed this has been reverted (for now) and Windows doesn't need it.
env_path_backup_var_exists = os.environ.get("CONDA_PATH_BACKUP", None)


@deprecated(
    "24.3",
    "24.5",
    addendum="Handled by `conda.gateways.connection.session.CondaSession`.",
)
def handle_proxy_407(x, y):
    pass


deprecated.constant(
    "24.3",
    "24.5",
    "hashsum_file",
    compute_sum,
    addendum="Use `conda.gateways.disk.read.compute_sum` instead.",
)


@deprecated(
    "24.3",
    "24.5",
    addendum="Use `conda.gateways.disk.read.compute_sum(path, 'md5')` instead.",
)
def md5_file(path: str | os.PathLike) -> str:
    return compute_sum(path, "md5")


@deprecated(
    "24.3",
    "24.5",
    addendum="Use `conda_build.environ.get_version_from_git_tag` instead.",
)
def get_version_from_git_tag(tag):
    from .environ import get_version_from_git_tag

    return get_version_from_git_tag(tag)



# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING

from .. import api
from ..utils import on_win
from . import validators as valid
from .main_render import get_render_parser

if TYPE_CHECKING:
    from argparse import ArgumentParser
    from typing import Sequence

logging.basicConfig(level=logging.INFO)


def get_parser() -> ArgumentParser:
    """Returns a parser object for this command"""
    p = get_render_parser()
    p.prog = "conda debug"
    p.description = """

Set up environments and activation scripts to debug your build or test phase.

"""
    # we do this one separately because we only allow one entry to conda render
    p.add_argument(
        "recipe_or_package_file_path",
        help=(
            "Path to recipe directory or package file to use for dependency and source information. "
            "If you use a recipe, you get the build/host env and source work directory.  If you use "
            "a package file, you get the test environments and the test_tmp folder."
        ),
        type=valid.validate_is_conda_pkg_or_recipe_dir,
    )
    p.add_argument(
        "-p",
        "--path",
        help=(
            "root path in which to place envs, source and activation script.  Defaults to a "
            "standard conda-build work folder (packagename_timestamp) in your conda-bld folder."
        ),
    )
    p.add_argument(
        "-o",
        "--output-id",
        help=(
            "fnmatch pattern that is associated with the output that you want to create an env for.  "
            "Must match only one file, as we don't support creating envs for more than one output at a time. "
            "The top-level recipe can be specified by passing 'TOPLEVEL' here"
        ),
    )
    p.add_argument(
        "-a",
        "--activate-string-only",
        action="store_true",
        help="Output only the string to the used generated activation script.  Use this for creating envs in scripted "
        "environments.",
    )

    # cut out some args from render that don't make sense here
    #    https://stackoverflow.com/a/32809642/1170370
    p._handle_conflict_resolve(
        None,
        [("--output", [_ for _ in p._actions if _.option_strings == ["--output"]][0])],
    )
    p._handle_conflict_resolve(
        None,
        [
            (
                "--bootstrap",
                [_ for _ in p._actions if _.option_strings == ["--bootstrap"]][0],
            )
        ],
    )
    p._handle_conflict_resolve(
        None,
        [
            (
                "--old-build-string",
                [_ for _ in p._actions if _.option_strings == ["--old-build-string"]][
                    0
                ],
            )
        ],
    )
    return p


def execute(args: Sequence[str] | None = None) -> int:
    parser = get_parser()
    parsed = parser.parse_args(args)

    try:
        activation_string = api.debug(
            parsed.recipe_or_package_file_path,
            verbose=(not parsed.activate_string_only),
            **parsed.__dict__,
        )

        if not parsed.activate_string_only:
            print("#" * 80)
            print(
                "Test environment created for debugging.  To enter a debugging environment:\n"
            )

        print(activation_string)
        if not parsed.activate_string_only:
            test_file = "conda_test_runner.bat" if on_win else "conda_test_runner.sh"
            print(
                f"To run your tests, you might want to start with running the {test_file} file."
            )
            print("#" * 80)

    except ValueError as e:
        print(
            f"Error: conda-debug encountered the following error:\n{e}", file=sys.stderr
        )
        sys.exit(1)

    return 0



# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
import os
import pprint
from os.path import dirname, isdir, isfile, join

# importing setuptools patches distutils so that it knows how to find VC for python 2.7
import setuptools  # noqa

# Leverage the hard work done by setuptools/distutils to find vcvarsall using
# either the registry or the VS**COMNTOOLS environment variable
try:
    from setuptools._distutils.msvc9compiler import WINSDK_BASE, Reg
    from setuptools._distutils.msvc9compiler import (
        find_vcvarsall as distutils_find_vcvarsall,
    )
except:
    # Allow some imports to work for cross or CONDA_SUBDIR usage.
    pass

from . import environ
from .utils import (
    check_call_env,
    copy_into,
    get_logger,
    path_prepended,
    write_bat_activation_text,
)
from .variants import get_default_variant, set_language_env_vars

VS_VERSION_STRING = {
    "8.0": "Visual Studio 8 2005",
    "9.0": "Visual Studio 9 2008",
    "10.0": "Visual Studio 10 2010",
    "11.0": "Visual Studio 11 2012",
    "12.0": "Visual Studio 12 2013",
    "14.0": "Visual Studio 14 2015",
}


def fix_staged_scripts(scripts_dir, config):
    """
    Fixes scripts which have been installed unix-style to have a .bat
    helper
    """
    if not isdir(scripts_dir):
        return
    for fn in os.listdir(scripts_dir):
        # process all the extensionless files
        if not isfile(join(scripts_dir, fn)) or "." in fn:
            continue

        # read as binary file to ensure we don't run into encoding errors, see #1632
        with open(join(scripts_dir, fn), "rb") as f:
            line = f.readline()
            # If it's a #!python script
            if not (line.startswith(b"#!") and b"python" in line.lower()):
                continue
            print(
                "Adjusting unix-style #! script %s, "
                "and adding a .bat file for it" % fn
            )
            # copy it with a .py extension (skipping that first #! line)
            with open(join(scripts_dir, fn + "-script.py"), "wb") as fo:
                fo.write(f.read())
            # now create the .exe file
            copy_into(
                join(dirname(__file__), "cli-%s.exe" % config.host_arch),
                join(scripts_dir, fn + ".exe"),
            )

        # remove the original script
        os.remove(join(scripts_dir, fn))


def build_vcvarsall_vs_path(version):
    """
    Given the Visual Studio version, returns the default path to the
    Microsoft Visual Studio vcvarsall.bat file.
    Expected versions are of the form {9.0, 10.0, 12.0, 14.0}
    """
    # Set up a load of paths that can be imported from the tests
    if "ProgramFiles(x86)" in os.environ:
        PROGRAM_FILES_PATH = os.environ["ProgramFiles(x86)"]
    else:
        PROGRAM_FILES_PATH = os.environ["ProgramFiles"]

    flatversion = str(version).replace(".", "")
    vstools = f"VS{flatversion}COMNTOOLS"

    if vstools in os.environ:
        return os.path.join(os.environ[vstools], "..\\..\\VC\\vcvarsall.bat")
    else:
        # prefer looking at env var; fall back to program files defaults
        return os.path.join(
            PROGRAM_FILES_PATH,
            f"Microsoft Visual Studio {version}",
            "VC",
            "vcvarsall.bat",
        )


def msvc_env_cmd(bits, config, override=None):
    # TODO: this function will likely break on `win-arm64`. However, unless
    # there's clear user demand, it's not clear that we should invest the
    # effort into updating a known deprecated function for a new platform.
    log = get_logger(__name__)
    log.warn(
        "Using legacy MSVC compiler setup.  This will be removed in conda-build 4.0. "
        "If this recipe does not use a compiler, this message is safe to ignore.  "
        "Otherwise, use {{compiler('<language>')}} jinja2 in requirements/build."
    )
    if bits not in ["64", "32"]:
        log.warn(f"The legacy MSVC compiler setup does not support {bits} builds. ")
        return ""
    if override:
        log.warn(
            "msvc_compiler key in meta.yaml is deprecated. Use the new"
            "variant-powered compiler configuration instead. Note that msvc_compiler"
            "is incompatible with the new {{{{compiler('c')}}}} jinja scheme."
        )
    # this has been an int at times.  Make sure it's a string for consistency.
    bits = str(bits)
    arch_selector = "x86" if bits == "32" else "amd64"

    msvc_env_lines = []

    version = None
    if override is not None:
        version = override

    # The DISTUTILS_USE_SDK variable tells distutils to not try and validate
    # the MSVC compiler. For < 3.5 this still forcibly looks for 'cl.exe'.
    # For > 3.5 it literally just skips the validation logic.
    # See distutils _msvccompiler.py and msvc9compiler.py / msvccompiler.py
    # for more information.
    msvc_env_lines.append("set DISTUTILS_USE_SDK=1")
    # This is also required to hit the 'don't validate' logic on < 3.5.
    # For > 3.5 this is ignored.
    msvc_env_lines.append("set MSSdk=1")

    if not version:
        py_ver = config.variant.get("python", get_default_variant(config)["python"])
        if int(py_ver[0]) >= 3:
            if int(py_ver.split(".")[1]) < 5:
                version = "10.0"
            version = "14.0"
        else:
            version = "9.0"

    if float(version) >= 14.0:
        # For Python 3.5+, ensure that we link with the dynamic runtime.  See
        # http://stevedower.id.au/blog/building-for-python-3-5-part-two/ for more info
        msvc_env_lines.append(
            "set PY_VCRUNTIME_REDIST=%LIBRARY_BIN%\\vcruntime{}.dll".format(
                version.replace(".", "")
            )
        )

    vcvarsall_vs_path = build_vcvarsall_vs_path(version)

    def build_vcvarsall_cmd(cmd, arch=arch_selector):
        # Default argument `arch_selector` is defined above
        return f'call "{cmd}" {arch}'

    vs_major = version.split(".")[0]
    msvc_env_lines.append(f'set "VS_VERSION={version}"')
    msvc_env_lines.append(f'set "VS_MAJOR={vs_major}"')
    msvc_env_lines.append(f'set "VS_YEAR={VS_VERSION_STRING[version][-4:]}"')
    if int(vs_major) >= 16:
        # No Win64 for VS 2019.
        msvc_env_lines.append(f'set "CMAKE_GENERATOR={VS_VERSION_STRING[version]}"')
    else:
        msvc_env_lines.append(
            'set "CMAKE_GENERATOR={}"'.format(
                VS_VERSION_STRING[version] + {"64": " Win64", "32": ""}[bits]
            )
        )
    # tell msys2 to ignore path conversions for issue-causing windows-style flags in build
    #   See https://github.com/conda-forge/icu-feedstock/pull/5
    msvc_env_lines.append('set "MSYS2_ARG_CONV_EXCL=/AI;/AL;/OUT;/out"')
    msvc_env_lines.append('set "MSYS2_ENV_CONV_EXCL=CL"')
    if version == "10.0":
        try:
            WIN_SDK_71_PATH = Reg.get_value(
                os.path.join(WINSDK_BASE, "v7.1"), "installationfolder"
            )
            WIN_SDK_71_BAT_PATH = os.path.join(WIN_SDK_71_PATH, "Bin", "SetEnv.cmd")

            win_sdk_arch = "/Release /x86" if bits == "32" else "/Release /x64"
            win_sdk_cmd = build_vcvarsall_cmd(WIN_SDK_71_BAT_PATH, arch=win_sdk_arch)

            # There are two methods of building Python 3.3 and 3.4 extensions (both
            # of which required Visual Studio 2010 - as explained in the Python wiki
            # https://wiki.python.org/moin/WindowsCompilers)
            # 1) Use the Windows SDK 7.1
            # 2) Use Visual Studio 2010 (any edition)
            # However, VS2010 never shipped with a 64-bit compiler, so in this case
            # **only** option (1) applies. For this reason, we always try and
            # activate the Windows SDK first. Unfortunately, unsuccessfully setting
            # up the environment does **not EXIT 1** and therefore we must fall
            # back to attempting to set up VS2010.
            # DelayedExpansion is required for the SetEnv.cmd
            msvc_env_lines.append("Setlocal EnableDelayedExpansion")
            msvc_env_lines.append(win_sdk_cmd)
            # If the WindowsSDKDir environment variable has not been successfully
            # set then try activating VS2010
            msvc_env_lines.append(
                f'if not "%WindowsSDKDir%" == "{WIN_SDK_71_PATH}" ( {build_vcvarsall_cmd(vcvarsall_vs_path)} )'
            )
        # sdk is not installed.  Fall back to only trying VS 2010
        except KeyError:
            msvc_env_lines.append(build_vcvarsall_cmd(vcvarsall_vs_path))
    elif version == "9.0":
        # Get the Visual Studio 2008 path (not the Visual C++ for Python path)
        # and get the 'vcvars64.bat' from inside the bin (in the directory above
        # that returned by distutils_find_vcvarsall)
        try:
            VCVARS64_VS9_BAT_PATH = os.path.join(
                os.path.dirname(distutils_find_vcvarsall(9)), "bin", "vcvars64.bat"
            )
        # there's an exception if VS or the VC compiler for python are not actually installed.
        except (KeyError, TypeError):
            VCVARS64_VS9_BAT_PATH = None

        error1 = "IF %ERRORLEVEL% NEQ 0 {}"

        # Prefer VS9 proper over Microsoft Visual C++ Compiler for Python 2.7
        msvc_env_lines.append(build_vcvarsall_cmd(vcvarsall_vs_path))
        # The Visual Studio 2008 Express edition does not properly contain
        # the amd64 build files, so we call the vcvars64.bat manually,
        # rather than using the vcvarsall.bat which would try and call the
        # missing bat file.
        if arch_selector == "amd64" and VCVARS64_VS9_BAT_PATH:
            msvc_env_lines.append(
                error1.format(build_vcvarsall_cmd(VCVARS64_VS9_BAT_PATH))
            )
        # Otherwise, fall back to icrosoft Visual C++ Compiler for Python 2.7+
        # by using the logic provided by setuptools
        msvc_env_lines.append(
            error1.format(build_vcvarsall_cmd(distutils_find_vcvarsall(9)))
        )
    else:
        # Visual Studio 14 or otherwise
        msvc_env_lines.append(build_vcvarsall_cmd(vcvarsall_vs_path))

    return "\n".join(msvc_env_lines) + "\n"


def write_build_scripts(m, env, bld_bat):
    env_script = join(m.config.work_dir, "build_env_setup.bat")
    if m.noarch == "python":
        env["PYTHONDONTWRITEBYTECODE"] = True
    import codecs

    with codecs.getwriter("utf-8")(open(env_script, "wb")) as fo:
        # more debuggable with echo on
        fo.write("@echo on\n")
        for key, value in env.items():
            if value != "" and value is not None:
                fo.write(f'set "{key}={value}"\n')
        if not m.uses_new_style_compiler_activation:
            fo.write(
                msvc_env_cmd(
                    bits=m.config.host_arch,
                    config=m.config,
                    override=m.get_value("build/msvc_compiler", None),
                )
            )
        # Reset echo on, because MSVC scripts might have turned it off
        fo.write("@echo on\n")
        fo.write('set "INCLUDE={};%INCLUDE%"\n'.format(env["LIBRARY_INC"]))
        fo.write('set "LIB={};%LIB%"\n'.format(env["LIBRARY_LIB"]))
        if m.config.activate and m.name() != "conda":
            write_bat_activation_text(fo, m)
    # bld_bat may have been generated elsewhere with contents of build/script
    work_script = join(m.config.work_dir, "conda_build.bat")
    if os.path.isfile(bld_bat):
        with open(bld_bat) as fi:
            data = fi.read()
        with codecs.getwriter("utf-8")(open(work_script, "wb")) as fo:
            fo.write('IF "%CONDA_BUILD%" == "" (\n')
            fo.write(f"    call {env_script}\n")
            fo.write(")\n")
            fo.write("REM ===== end generated header =====\n")
            fo.write(data)
    return work_script, env_script


def build(m, bld_bat, stats, provision_only=False):
    # TODO: Prepending the prefixes here should probably be guarded by
    #         if not m.activate_build_script:
    #       Leaving it as is, for now, since we need a quick, non-disruptive patch release.
    with path_prepended(m.config.host_prefix):
        with path_prepended(m.config.build_prefix):
            env = environ.get_dict(m=m)
    env["CONDA_BUILD_STATE"] = "BUILD"

    # hard-code this because we never want pip's build isolation
    #    https://github.com/conda/conda-build/pull/2972#discussion_r198290241
    #
    # Note that pip env "NO" variables are inverted logic.
    #      PIP_NO_BUILD_ISOLATION=False means don't use build isolation.
    #
    env["PIP_NO_BUILD_ISOLATION"] = "False"
    # some other env vars to have pip ignore dependencies.
    # we supply them ourselves instead.
    #    See note above about inverted logic on "NO" variables
    env["PIP_NO_DEPENDENCIES"] = True
    env["PIP_IGNORE_INSTALLED"] = True

    # pip's cache directory (PIP_NO_CACHE_DIR) should not be
    # disabled as this results in .egg-info rather than
    # .dist-info directories being created, see gh-3094
    # set PIP_CACHE_DIR to a path in the work dir that does not exist.
    env["PIP_CACHE_DIR"] = m.config.pip_cache_dir

    # tell pip to not get anything from PyPI, please.  We have everything we need
    # locally, and if we don't, it's a problem.
    env["PIP_NO_INDEX"] = True

    # set variables like CONDA_PY in the test environment
    env.update(set_language_env_vars(m.config.variant))

    for name in "BIN", "INC", "LIB":
        path = env["LIBRARY_" + name]
        if not isdir(path):
            os.makedirs(path)

    work_script, env_script = write_build_scripts(m, env, bld_bat)

    if not provision_only and os.path.isfile(work_script):
        cmd = ["cmd.exe", "/d", "/c", os.path.basename(work_script)]
        # rewrite long paths in stdout back to their env variables
        if m.config.debug or m.config.no_rewrite_stdout_env:
            rewrite_env = None
        else:
            rewrite_env = {
                k: env[k] for k in ["PREFIX", "BUILD_PREFIX", "SRC_DIR"] if k in env
            }
            print("Rewriting env in output: %s" % pprint.pformat(rewrite_env))
        check_call_env(
            cmd, cwd=m.config.work_dir, stats=stats, rewrite_stdout_env=rewrite_env
        )
        fix_staged_scripts(join(m.config.host_prefix, "Scripts"), config=m.config)



# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
"""
Tools for converting PyPI packages to conda recipes.
"""
import keyword
import logging
import os
import re
import subprocess
import sys
from collections import OrderedDict, defaultdict
from os import chdir, getcwd, listdir, makedirs
from os.path import abspath, exists, isdir, isfile, join
from shutil import copy2
from tempfile import mkdtemp
from urllib.parse import urljoin, urlsplit

import pkginfo
import requests
import yaml
from conda.gateways.disk.read import compute_sum
from requests.packages.urllib3.util.url import parse_url

from ..conda_interface import (
    StringIO,
    configparser,
    default_python,
    download,
    human_bytes,
    input,
    normalized_version,
    spec_from_line,
)
from ..config import Config
from ..environ import create_env
from ..license_family import allowed_license_families, guess_license_family
from ..metadata import MetaData
from ..render import FIELDS as EXPECTED_SECTION_ORDER
from ..source import apply_patch
from ..utils import (
    check_call_env,
    decompressible_exts,
    ensure_list,
    on_win,
    rm_rf,
    tar_xf,
)
from ..version import _parse as parse_version

pypi_example = """
Examples:

Create a recipe for the sympy package:

    conda skeleton pypi sympy

Create a recipes for the flake8 package and all its dependencies:

    conda skeleton pypi --recursive flake8

Use the --pypi-url flag to point to a PyPI mirror url:

    conda skeleton pypi --pypi-url <mirror-url> package_name
"""

# Definition of REQUIREMENTS_ORDER below are from
# https://github.com/conda-forge/conda-smithy/blob/master/conda_smithy/lint_recipe.py#L16
REQUIREMENTS_ORDER = ["host", "run"]

# Definition of ABOUT_ORDER reflects current practice
ABOUT_ORDER = [
    "home",
    "license",
    "license_family",
    "license_file",
    "summary",
    "doc_url",
    "dev_url",
]

PYPI_META_HEADER = """{{% set name = "{packagename}" %}}
{{% set version = "{version}" %}}

"""

# To preserve order of the output in each of the sections the data type
# needs to be ordered.
# The top-level ordering is irrelevant because the write order of 'package',
# etc. is determined by EXPECTED_SECTION_ORDER.
PYPI_META_STATIC = {
    "package": OrderedDict(
        [
            ("name", "{{ name|lower }}"),
            ("version", "{{ version }}"),
        ]
    ),
    "source": OrderedDict(
        [
            (
                "url",
                "/packages/source/{{ name[0] }}/{{ name }}/{{ name }}-{{ version }}.tar.gz",
            ),  # NOQA
        ]
    ),
    "build": OrderedDict(
        [
            ("number", 0),
        ]
    ),
    "extra": OrderedDict([("recipe-maintainers", "")]),
}


# Note the {} formatting bits here
DISTUTILS_PATCH = '''\
diff core.py core.py
--- core.py
+++ core.py
@@ -166,5 +167,40 @@ def setup (**attrs):
 \n
+# ====== BEGIN CONDA SKELETON PYPI PATCH ======
+
+import distutils.core
+import io
+import os.path
+import sys
+import ruamel.yaml as yaml
+from ruamel.yaml import Loader, SafeLoader
+
+# Override the default string handling function to always return unicode
+# objects (taken from StackOverflow)
+def construct_yaml_str(self, node):
+    return self.construct_scalar(node)
+Loader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
+SafeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)
+
+def setup(*args, **kwargs):
+    data = {{}}
+    data['tests_require'] = kwargs.get('tests_require', [])
+    data['install_requires'] = kwargs.get('install_requires', [])
+    data['extras_require'] = kwargs.get('extras_require', {{}})
+    data['entry_points'] = kwargs.get('entry_points', [])
+    data['packages'] = kwargs.get('packages', [])
+    data['setuptools'] = 'setuptools' in sys.modules
+    data['summary'] = kwargs.get('description', None)
+    data['home'] = kwargs.get('url', None)
+    data['license'] = kwargs.get('license', None)
+    data['name'] = kwargs.get('name', '??PACKAGE-NAME-UNKNOWN??')
+    data['classifiers'] = kwargs.get('classifiers', None)
+    data['version'] = kwargs.get('version', '??PACKAGE-VERSION-UNKNOWN??')
+    with io.open(os.path.join("{}", "pkginfo.yaml"), 'w', encoding='utf-8') as fn:
+        fn.write(yaml.safe_dump(data, encoding=None))
+
+
+# ======= END CONDA SKELETON PYPI PATCH ======
 \n
 def run_setup (script_name, script_args=None, stop_after="run"):
     """Run a setup script in a somewhat controlled environment, and
'''

INDENT = "\n    - "


def _ssl_no_verify():
    """Gets whether the SSL_NO_VERIFY environment variable is set to 1 or True.

    This provides a workaround for users in some corporate environments where
    MITM style proxies make it difficult to fetch data over HTTPS.
    """
    return os.environ.get("SSL_NO_VERIFY", "").strip().lower() in ("1", "true")


def package_exists(package_name, pypi_url=None):
    if not pypi_url:
        pypi_url = "https://pypi.io/pypi"
    # request code will be 404 if the package does not exist.  Requires exact match.
    r = requests.get(pypi_url + "/" + package_name, verify=not _ssl_no_verify())
    return r.status_code != 404


def __print_with_indent(line, prefix="", suffix="", level=0, newline=True):
    output = ""
    if level:
        output = " " * level
    return output + prefix + line + suffix + ("\n" if newline else "")


def _print_dict(recipe_metadata, order=None, level=0, indent=2):
    """Free function responsible to get the metadata which represents the
    recipe and convert it to the yaml format.

    :param OrderedDict recipe_metadata:
    :param list order: Order to be write each section
    :param int level:
    :param int indent: Indentation - Number of empty spaces for each level
    :return string: Recipe rendered with the metadata
    """
    rendered_recipe = ""
    if not order:
        order = sorted(list(recipe_metadata.keys()))
    for section_name in order:
        if section_name in recipe_metadata and recipe_metadata[section_name]:
            rendered_recipe += __print_with_indent(section_name, suffix=":")
            for attribute_name, attribute_value in recipe_metadata[
                section_name
            ].items():
                if attribute_value is None:
                    continue
                if isinstance(attribute_value, str) or not hasattr(
                    attribute_value, "__iter__"
                ):
                    rendered_recipe += __print_with_indent(
                        attribute_name, suffix=":", level=level + indent, newline=False
                    )
                    rendered_recipe += _formating_value(attribute_name, attribute_value)
                elif hasattr(attribute_value, "keys"):
                    rendered_recipe += _print_dict(
                        attribute_value, sorted(list(attribute_value.keys()))
                    )
                # assume that it's a list if it exists at all
                elif attribute_value:
                    rendered_recipe += __print_with_indent(
                        attribute_name, suffix=":", level=level + indent
                    )
                    for item in attribute_value:
                        rendered_recipe += __print_with_indent(
                            item, prefix="- ", level=level + indent
                        )
            # add a newline in between sections
            if level == 0:
                rendered_recipe += "\n"

    return rendered_recipe


def _formating_value(attribute_name, attribute_value):
    """Format the value of the yaml file. This function will quote the
    attribute value if needed.

    :param string attribute_name: Attribute name
    :param string attribute_value: Attribute value
    :return string: Value quoted if need
    """
    pattern_search = re.compile(r"[@_!#$%^&*()<>?/\|}{~:]")
    if (
        isinstance(attribute_value, str)
        and pattern_search.search(attribute_value)
        or attribute_name in ["summary", "description", "version", "script"]
    ):
        return ' "' + str(attribute_value) + '"\n'
    return " " + str(attribute_value) + "\n"


def skeletonize(
    packages,
    output_dir=".",
    version=None,
    recursive=False,
    all_urls=False,
    pypi_url="https://pypi.io/pypi/",
    noprompt=True,
    version_compare=False,
    python_version=None,
    manual_url=False,
    all_extras=False,
    noarch_python=False,
    config=None,
    setup_options=None,
    extra_specs=[],
    pin_numpy=False,
):
    package_dicts = {}

    if not setup_options:
        setup_options = []

    if isinstance(setup_options, str):
        setup_options = [setup_options]

    if not config:
        config = Config()

    python_version = python_version or config.variant.get("python", default_python)

    created_recipes = []
    while packages:
        package = packages.pop()
        created_recipes.append(package)

        is_url = ":" in package

        if is_url:
            package_pypi_url = ""
        else:
            package_pypi_url = urljoin(pypi_url, "/".join((package, "json")))

        if not is_url:
            dir_path = join(output_dir, package.lower())
            if exists(dir_path) and not version_compare:
                raise RuntimeError("directory already exists: %s" % dir_path)
        d = package_dicts.setdefault(
            package,
            {
                "packagename": package,
                "run_depends": "",
                "build_depends": "",
                "entry_points": "",
                "test_commands": "",
                "tests_require": "",
            },
        )
        if is_url:
            del d["packagename"]

        if is_url:
            d["version"] = "UNKNOWN"
            # Make sure there is always something to pass in for this
            pypi_data = {}
        else:
            pypi_resp = requests.get(package_pypi_url, verify=not _ssl_no_verify())

            if pypi_resp.status_code != 200:
                sys.exit(
                    "Request to fetch %s failed with status: %d"
                    % (package_pypi_url, pypi_resp.status_code)
                )

            pypi_data = pypi_resp.json()

            versions = sorted(pypi_data["releases"].keys(), key=parse_version)

            if version_compare:
                version_compare(versions)
            if version:
                if version not in versions:
                    sys.exit(
                        f"Error: Version {version} of {package} is not available on PyPI."
                    )
                d["version"] = version
            else:
                # select the most visible version from PyPI.
                if not versions:
                    sys.exit(
                        "Error: Could not find any versions of package %s" % package
                    )
                if len(versions) > 1:
                    print("Warning, the following versions were found for %s" % package)
                    for ver in versions:
                        print(ver)
                    print("Using %s" % versions[-1])
                    print("Use --version to specify a different version.")
                d["version"] = versions[-1]

        data, d["pypiurl"], d["filename"], d["digest"] = get_download_data(
            pypi_data, package, d["version"], is_url, all_urls, noprompt, manual_url
        )

        d["import_tests"] = ""

        # Get summary directly from the metadata returned
        # from PyPI. summary will be pulled from package information in
        # get_package_metadata or a default value set if it turns out that
        # data['summary'] is empty.  Ignore description as it is too long.
        d["summary"] = data.get("summary", "")
        get_package_metadata(
            package,
            d,
            data,
            output_dir,
            python_version,
            all_extras,
            recursive,
            created_recipes,
            noarch_python,
            noprompt,
            packages,
            extra_specs,
            config=config,
            setup_options=setup_options,
        )

        # Set these *after* get_package_metadata so that the preferred hash
        # can be calculated from the downloaded file, if necessary.
        d["hash_type"] = d["digest"][0]
        d["hash_value"] = d["digest"][1]

        # Change requirements to use format that guarantees the numpy
        # version will be pinned when the recipe is built and that
        # the version is included in the build string.
        if pin_numpy:
            for depends in ["build_depends", "run_depends"]:
                deps = d[depends]
                numpy_dep = [idx for idx, dep in enumerate(deps) if "numpy" in dep]
                if numpy_dep:
                    # Turns out this needs to be inserted before the rest
                    # of the numpy spec.
                    deps.insert(numpy_dep[0], "numpy x.x")
                    d[depends] = deps

    for package in package_dicts:
        d = package_dicts[package]
        name = d["packagename"].lower()
        makedirs(join(output_dir, name))
        print("Writing recipe for %s" % package.lower())
        with open(join(output_dir, name, "meta.yaml"), "w") as f:
            rendered_recipe = PYPI_META_HEADER.format(**d)

            ordered_recipe = OrderedDict()
            # Create all keys in expected ordered
            for key in EXPECTED_SECTION_ORDER:
                try:
                    ordered_recipe[key] = PYPI_META_STATIC[key]
                except KeyError:
                    ordered_recipe[key] = OrderedDict()

            if "://" not in pypi_url:
                raise ValueError("pypi_url must have protocol (e.g. http://) included")
            base_url = urlsplit(pypi_url)
            base_url = "://".join((base_url.scheme, base_url.netloc))
            ordered_recipe["source"]["url"] = urljoin(
                base_url, ordered_recipe["source"]["url"]
            )
            ordered_recipe["source"]["sha256"] = d["hash_value"]

            if d["entry_points"]:
                ordered_recipe["build"]["entry_points"] = d["entry_points"]

            if noarch_python:
                ordered_recipe["build"]["noarch"] = "python"

            recipe_script_cmd = [
                "{{ PYTHON }} -m pip install . -vv --no-deps --no-build-isolation"
            ]
            ordered_recipe["build"]["script"] = " ".join(
                recipe_script_cmd + setup_options
            )

            # Always require python as a dependency.  Pip is because we use pip for
            #    the install line.
            ordered_recipe["requirements"] = OrderedDict()
            ordered_recipe["requirements"]["host"] = sorted(
                set(["python", "pip"] + list(d["build_depends"]))
            )
            ordered_recipe["requirements"]["run"] = sorted(
                set(["python"] + list(d["run_depends"]))
            )

            if d["import_tests"]:
                ordered_recipe["test"]["imports"] = d["import_tests"]

            if d["test_commands"]:
                ordered_recipe["test"]["commands"] = d["test_commands"]

            if d["tests_require"]:
                ordered_recipe["test"]["requires"] = d["tests_require"]

            ordered_recipe["about"] = OrderedDict()

            for key in ABOUT_ORDER:
                try:
                    ordered_recipe["about"][key] = d[key]
                except KeyError:
                    ordered_recipe["about"][key] = ""
            ordered_recipe["extra"]["recipe-maintainers"] = ["your-github-id-here"]

            # Prune any top-level sections that are empty
            rendered_recipe += _print_dict(ordered_recipe, EXPECTED_SECTION_ORDER)

            # make sure that recipe ends with one newline, by god.
            rendered_recipe.rstrip()

            # This hackery is necessary because
            #  - the default indentation of lists is not what we would like.
            #    Ideally we'd contact the ruamel.yaml author to find the right
            #    way to do this. See this PR thread for more:
            #    https://github.com/conda/conda-build/pull/2205#issuecomment-315803714
            #    Brute force fix below.

            # Fix the indents
            recipe_lines = []
            for line in rendered_recipe.splitlines():
                match = re.search(r"^\s+(-) ", line, flags=re.MULTILINE)
                if match:
                    pre, sep, post = line.partition("-")
                    sep = "  " + sep
                    line = pre + sep + post
                recipe_lines.append(line)
            rendered_recipe = "\n".join(recipe_lines)

            f.write(rendered_recipe)


def add_parser(repos):
    """Modify repos in place, adding the PyPI option"""
    pypi = repos.add_parser(
        "pypi",
        help="""
    Create recipe skeleton for packages hosted on the Python Packaging Index
    (PyPI) (pypi.io).
        """,
        epilog=pypi_example,
    )
    pypi.add_argument(
        "packages",
        nargs="+",
        help="""PyPi packages to create recipe skeletons for.
                You can also specify package[extra,...] features.""",
    )
    pypi.add_argument(
        "--output-dir",
        help="Directory to write recipes to (default: %(default)s).",
        default=".",
    )
    pypi.add_argument(
        "--version",
        help="""Version to use. Applies to all packages. If not specified the
              lastest visible version on PyPI is used.""",
    )
    pypi.add_argument(
        "--all-urls",
        action="store_true",
        help="""Look at all URLs, not just source URLs. Use this if it can't
                find the right URL.""",
    )
    pypi.add_argument(
        "--pypi-url",
        default="https://pypi.io/pypi/",
        help="URL to use for PyPI (default: %(default)s).",
    )
    pypi.add_argument(
        "--prompt",
        action="store_false",
        default=True,
        dest="noprompt",
        help="""Prompt the user on ambiguous choices.  Default is to make the
        best possible choice and continue.""",
    )
    pypi.add_argument(
        "--all-extras",
        action="store_true",
        default=False,
        help="Add all extra feature requirements. Applies to all packages.",
    )
    pypi.add_argument(
        "--recursive",
        action="store_true",
        help="Create recipes for dependencies if they do not already exist.",
    )
    pypi.add_argument(
        "--version-compare",
        action="store_true",
        help="""Compare the package version of the recipe with all available
        versions on PyPI.""",
    )
    pypi.add_argument(
        "--python-version",
        action="store",
        default=default_python,
        help="""Version of Python to use to run setup.py. Default is %(default)s.""",
        choices=["2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10", "3.11", "3.12"],
    )

    pypi.add_argument(
        "--manual-url",
        action="store_true",
        default=False,
        help=(
            "Manually choose source url when more than one urls are present."
            "Default is the one with least source size."
        ),
    )

    pypi.add_argument(
        "--noarch-python",
        action="store_true",
        default=False,
        help="Creates recipe as noarch python",
    )

    pypi.add_argument(
        "--setup-options",
        action="append",
        default=[],
        help="Options to be added to setup.py install in the recipe. "
        "The same options are passed to setup.py install in both "
        "the construction of the recipe and in the recipe itself."
        "For options that include a double-hypen or to pass multiple "
        "options, use the syntax "
        '--setup-options="--option1 --option-with-arg arg"',
    )

    pypi.add_argument(
        "--pin-numpy",
        action="store_true",
        help="Ensure that the generated recipe pins the version of numpy"
        "to CONDA_NPY.",
    )

    pypi.add_argument(
        "--extra-specs",
        action="append",
        default=[],
        help="Extra specs for the build environment to extract the skeleton.",
    )


def get_download_data(
    pypi_data, package, version, is_url, all_urls, noprompt, manual_url
):
    """
    Get at least one valid *source* download URL or fail.

    Returns
    -------

    data : dict
        Summary of package information
    pypiurl : str
        Download URL of package, which may or may not actually be from PyPI.
    filename : str
        Name of file; used to check cache
    digest : dict
        Key is type of checksum, value is the checksum.
    """
    data = pypi_data["info"] if not is_url else {}

    # PyPI will typically have several downloads (source, wheels) for one
    # package/version.
    urls = [url for url in pypi_data["releases"][version]] if not is_url else [package]

    if not is_url and not all_urls:
        # Try to find source urls
        urls = [url for url in urls if url["packagetype"] == "sdist"]

    if not urls:
        # Try harder for a download location
        if data.get("download_url"):
            urls = [defaultdict(str, {"url": data["download_url"]})]
            if not urls[0]["url"]:
                # The package doesn't have a url, or maybe it only has a wheel.
                sys.exit(
                    "Error: Could not build recipe for %s. "
                    "Could not find any valid urls." % package
                )
            U = parse_url(urls[0]["url"])
            if not U.path:
                sys.exit(f"Error: Could not parse url for {package}: {U}")
            urls[0]["filename"] = U.path.rsplit("/")[-1]
            fragment = U.fragment or ""
            digest = fragment.split("=")
        else:
            sys.exit("Error: No source urls found for %s" % package)
    if len(urls) > 1 and not noprompt:
        print("More than one source version is available for %s:" % package)
        if manual_url:
            for i, url in enumerate(urls):
                print(
                    "%d: %s (%s) %s"
                    % (i, url["url"], human_bytes(url["size"]), url["comment_text"])
                )
            n = int(input("which version should i use? "))
        else:
            print("Using the one with the least source size")
            print("use --manual-url to override this behavior")
            _, n = min((url["size"], i) for (i, url) in enumerate(urls))
    else:
        n = 0

    if not is_url:
        # Found a location from PyPI.
        url = urls[n]
        pypiurl = url["url"]
        print(
            "Using url {} ({}) for {}.".format(
                pypiurl, human_bytes(url["size"] or 0), package
            )
        )

        if url["digests"]["sha256"]:
            digest = ("sha256", url["digests"]["sha256"])
        else:
            # That didn't work, even though as of 7/17/2017 some packages
            # have a 'digests' entry.
            # As a last-ditch effort, try for the md5_digest entry.
            digest = ()
        filename = url["filename"] or "package"
    else:
        # User provided a URL, try to use it.
        print("Using url %s" % package)
        pypiurl = package
        U = parse_url(package)
        digest = U.fragment.split("=")
        # TODO: 'package' won't work with unpack()
        filename = U.path.rsplit("/", 1)[-1] or "package"

    return (data, pypiurl, filename, digest)


def version_compare(package, versions):
    if not versions:
        # PyPI is case sensitive, this will pass control
        # to a method in main() to take care of that.
        return

    nv = normalized_version

    norm_versions = [nv(ver) for ver in versions]

    recipe_dir = abspath(package.lower())
    if not isdir(recipe_dir):
        sys.exit("Error: no such directory: %s" % recipe_dir)
    m = MetaData(recipe_dir)
    local_version = nv(m.version())
    print(f"Local recipe for {package} has version {local_version}")
    if local_version not in versions:
        sys.exit(f"Error: {package} {local_version} is not available on PyPI.")
    else:
        # Comparing normalized versions, displaying non normalized ones
        new_versions = versions[: norm_versions.index(local_version)]
        if len(new_versions) > 0:
            print("Following new versions of %s are avaliable" % (package))
            for ver in new_versions:
                print(ver)
        else:
            print("No new version for %s is available" % (package))
        sys.exit()


def convert_version(version):
    """Convert version into a pin-compatible format according to PEP440."""
    version_parts = version.split(".")
    suffixes = ("post", "pre")
    if any(suffix in version_parts[-1] for suffix in suffixes):
        version_parts.pop()
    # the max pin length is n-1, but in terms of index this is n-2
    max_ver_len = len(version_parts) - 2
    version_parts[max_ver_len] = int(version_parts[max_ver_len]) + 1
    max_pin = ".".join(str(v) for v in version_parts[: max_ver_len + 1])
    pin_compatible = f" >={version},<{max_pin}"
    return pin_compatible


MARKER_RE = re.compile(
    r"(?P<name>^[^=<>!~\s;]+)"
    r"\s*"
    r"(?P<constraint>[=!><~]=?\s*[^\s;]+)?"
    r"(?:\s*;\s+)?(?P<env_mark_name>[^=<>!~\s;]+)?"
    r"\s*"
    r"(?P<env_mark_constraint>[=<>!\s]+[^=<>!~\s]+)?"
)


def _get_env_marker_operator_and_value(constraint):
    operator, value = constraint.split()
    value = value.strip("'").strip('"').strip()
    return operator, value


def _translate_python_constraint(constraint):
    translation = constraint
    if len(constraint.split()) == 2:
        operator, value = _get_env_marker_operator_and_value(constraint)
        value = "".join(value.split(".")[:2])
        translation = " ".join((operator, value))
    return f"py {translation}"


def _translate_platform_system_constraint(constraint):
    operator, value = _get_env_marker_operator_and_value(constraint)
    system = {
        "Linux": "linux",
        "Darwin": "osx",
        "Windows": "win",
    }.get(value, value.lower())
    return "{}{}".format("not " if operator == "!=" else "", system)


def _translate_sys_platform_constraint(constraint):
    operator, value = _get_env_marker_operator_and_value(constraint)
    # Only take the "letter" part to translate, e.g., "linux2"->"linux", "win32"->"win".
    system = re.match("^[a-z]*", value, re.I)[0]
    return "{}{}".format("not " if operator == "!=" else "", system)


def env_mark_lookup(env_mark_name, env_mark_constraint):
    """returns translated variable name and corresponding function to run to normalize the
    version constraint to conda style"""
    # TODO: implement more of these from PEP 508 as necessary:
    #   https://www.python.org/dev/peps/pep-0508/
    env_mark_table = {
        "python_version": _translate_python_constraint,
        "platform_system": _translate_platform_system_constraint,
        "sys_platform": _translate_sys_platform_constraint,
    }
    marker = env_mark_table[env_mark_name](env_mark_constraint)
    return "  # [ " + marker + " ]"


def parse_dep_with_env_marker(dep_str):
    match = MARKER_RE.match(dep_str)
    name = match.group("name")
    if match.group("constraint"):
        name = " ".join((name, match.group("constraint").replace(" ", "")))
    env_mark = ""
    if match.group("env_mark_name"):
        env_mark = env_mark_lookup(
            match.group("env_mark_name"), match.group("env_mark_constraint")
        )
    return name, env_mark


def get_package_metadata(
    package,
    metadata,
    data,
    output_dir,
    python_version,
    all_extras,
    recursive,
    created_recipes,
    noarch_python,
    no_prompt,
    packages,
    extra_specs,
    config,
    setup_options,
):
    print("Downloading %s" % package)
    print("PyPI URL: ", metadata["pypiurl"])
    pkginfo = get_pkginfo(
        package,
        filename=metadata["filename"],
        pypiurl=metadata["pypiurl"],
        digest=metadata["digest"],
        python_version=python_version,
        extra_specs=extra_specs,
        setup_options=setup_options,
        config=config,
    )

    metadata.update(get_entry_points(pkginfo))

    requires = get_requirements(package, pkginfo, all_extras=all_extras)

    if requires or is_setuptools_enabled(pkginfo):
        list_deps = get_dependencies(requires, is_setuptools_enabled(pkginfo))

        metadata["build_depends"] = ["pip"] + list_deps
        # Never add setuptools to runtime dependencies.
        metadata["run_depends"] = list_deps

        if recursive:
            packages += get_recursive_deps(created_recipes, list_deps, output_dir)

    if "packagename" not in metadata:
        metadata["packagename"] = pkginfo["name"].lower()

    if metadata["version"] == "UNKNOWN":
        metadata["version"] = pkginfo["version"]

    metadata["import_tests"] = get_import_tests(pkginfo, metadata.get("import_tests"))
    metadata["tests_require"] = get_tests_require(pkginfo)

    metadata["home"] = get_home(pkginfo, data)

    if not metadata.get("summary"):
        metadata["summary"] = get_summary(pkginfo)

    license_name = get_license_name(package, pkginfo, no_prompt, data)
    metadata["license"] = clean_license_name(license_name)
    metadata["license_family"] = guess_license_family(
        license_name, allowed_license_families
    )

    if "new_hash_value" in pkginfo:
        metadata["digest"] = pkginfo["new_hash_value"]


def get_recursive_deps(created_recipes, list_deps, output_dir):
    """Function responsible to return the list of dependencies of the other
    projects which were requested.

    :param list created_recipes:
    :param list list_deps:
    :param output_dir:
    :return list:
    """
    recursive_deps = []
    for dep in list_deps:
        dep = dep.split()[0]
        if exists(join(output_dir, dep)) or dep in created_recipes:
            continue
        recursive_deps.append(dep)

    return recursive_deps


def get_dependencies(requires, setuptools_enabled=True):
    """Return the whole dependencies of the specified package
    :param list requires: List of requirements
    :param Bool setuptools_enabled: True if setuptools is enabled and False otherwise
    :return list: Return list of dependencies
    """

    # START :: Copied from conda
    # These can be removed if we want to drop support for conda <= 4.9.0
    def _strip_comment(line):
        return line.split("#")[0].rstrip()

    def _spec_from_line(line):
        spec_pat = re.compile(
            r"""
            (?P<name>[^=<>!\s]+)  # package name
            \s*  # ignore spaces
            (
                (?P<cc>=[^=]+(=[^=]+)?)  # conda constraint
                |
                (?P<pc>(?:[=!]=|[><]=?|~=).+)  # new pip-style constraints
            )?$
            """,
            re.VERBOSE,
        )
        m = spec_pat.match(_strip_comment(line))
        if m is None:
            return None
        name, cc, pc = (m.group("name").lower(), m.group("cc"), m.group("pc"))
        if cc:
            return name + cc.replace("=", " ")
        elif pc:
            if pc.startswith("~= "):
                assert (
                    pc.count("~=") == 1
                ), f"Overly complex 'Compatible release' spec not handled {line}"
                assert pc.count("."), f"No '.' in 'Compatible release' version {line}"
                ver = pc.replace("~= ", "")
                ver2 = ".".join(ver.split(".")[:-1]) + ".*"
                return name + " >=" + ver + ",==" + ver2
            else:
                return name + " " + pc.replace(" ", "")
        else:
            return name

    # END :: Copied from conda

    list_deps = ["setuptools"] if setuptools_enabled else []

    for dep_text in requires:
        if isinstance(dep_text, str):
            dep_text = dep_text.splitlines()
        # Every item may be a single requirement
        #  or a multiline requirements string...
        for dep in dep_text:
            # ... and may also contain comments...
            dep = dep.split("#")[0].strip()
            if not dep:
                continue

            dep_orig = dep
            dep, marker = parse_dep_with_env_marker(dep_orig)
            # if '~' in dep_orig:
            #     # dep_orig = 'kubernetes ~= 11'
            #     version = dep_orig.split()[-1]
            #     tilde_version = '~= {}'.format(version)
            #     pin_compatible = convert_version(version)
            #     spec = dep_orig.replace(tilde_version, pin_compatible)
            #     spec2 = spec_from_line(dep)
            #     if spec != spec2:
            #         print("Disagreement on PEP440 'Compatible release' {} vs {}".format(spec, spec2))
            spec = spec_from_line(dep)
            if "~=" in dep_orig:
                spec = None

            if spec is None:
                if "~=" in dep_orig:
                    log = logging.getLogger(__name__)
                    log.warning(
                        "Your conda is too old to handle ~= PEP440 'Compatible versions', "
                        "using copied implementation."
                    )
                    spec = _spec_from_line(dep_orig)
                if spec is None:
                    sys.exit("Error: Could not parse: %s" % dep)

            if marker:
                spec = " ".join((spec, marker))

            list_deps.append(spec)
    return list_deps


def get_import_tests(pkginfo, import_tests_metada=""):
    """Return the section import in tests

    :param dict pkginfo: Package information
    :param dict import_tests_metada: Imports already present
    :return list: Sorted list with the libraries necessary for the test phase
    """
    if not pkginfo.get("packages"):
        return import_tests_metada

    olddeps = []
    if import_tests_metada != "PLACEHOLDER":
        olddeps = [x for x in import_tests_metada.split() if x != "-"]
    return sorted(set(olddeps) | set(pkginfo["packages"]))


def get_tests_require(pkginfo):
    return sorted(spec_from_line(pkg) for pkg in ensure_list(pkginfo["tests_require"]))


def get_home(pkginfo, data=None):
    default_home = "The package home page"
    if pkginfo.get("home"):
        return pkginfo["home"]
    if data:
        return data.get("home", default_home)
    return default_home


def get_summary(pkginfo):
    return pkginfo.get("summary", "Summary of the package").replace('"', r"\"")


def get_license_name(package, pkginfo, no_prompt=False, data=None):
    """Responsible to return the license name
    :param str package: Package's name
    :param dict pkginfo:Package information
    :param no_prompt: If Prompt is not enabled
    :param dict data: Data
    :return str: License name
    """
    license_classifier = "License :: OSI Approved :: "

    data_classifier = data.get("classifiers", []) if data else []
    pkg_classifier = pkginfo.get("classifiers", data_classifier)
    pkg_classifier = pkg_classifier if pkg_classifier else data_classifier

    licenses = [
        classifier.split(license_classifier, 1)[1]
        for classifier in pkg_classifier
        if classifier.startswith(license_classifier)
    ]

    if licenses:
        return " or ".join(licenses)

    if pkginfo.get("license"):
        license_name = pkginfo["license"]
    elif data and "license" in data:
        license_name = data["license"]
    else:
        license_name = None

    if license_name:
        if no_prompt:
            return license_name
        elif "\n" not in license_name:
            print('Using "%s" for the license' % license_name)
        else:
            # Some projects put the whole license text in this field
            print("This is the license for %s" % package)
            print()
            print(license_name)
            print()
            license_name = input("What license string should I use? ")
    elif no_prompt:
        license_name = "UNKNOWN"
    else:
        license_name = input(
            "No license could be found for %s on PyPI or in the source. "
            "What license should I use? " % package
        )
    return license_name


def clean_license_name(license_name):
    """Remove the word ``license`` from the license
    :param str license_name: Receives the license name
    :return str: Return a string without the word ``license``
    """
    return re.subn(r"(.*)\s+license", r"\1", license_name, flags=re.IGNORECASE)[0]


def get_entry_points(pkginfo):
    """Look at the entry_points and construct console_script and gui_scripts entry_points for conda
    :param pkginfo:
    :return dict:
    """
    entry_points = pkginfo.get("entry_points")
    if not entry_points:
        return {}

    if isinstance(entry_points, str):
        # makes sure it is left-shifted
        newstr = "\n".join(x.strip() for x in entry_points.splitlines())
        _config = configparser.ConfigParser()

        try:
            _config.read_file(StringIO(newstr))
        except Exception as err:
            print("WARNING: entry-points not understood: ", err)
            print("The string was", newstr)
        else:
            entry_points = {}
            for section in _config.sections():
                if section in ["console_scripts", "gui_scripts"]:
                    entry_points[section] = [
                        f"{option}={_config.get(section, option)}"
                        for option in _config.options(section)
                    ]

    if isinstance(entry_points, dict):
        console_script = convert_to_flat_list(entry_points.get("console_scripts", []))
        gui_scripts = convert_to_flat_list(entry_points.get("gui_scripts", []))

        # TODO: Use pythonw for gui scripts
        entry_list = console_script + gui_scripts
        if entry_list:
            return {
                "entry_points": entry_list,
                "test_commands": make_entry_tests(entry_list),
            }
    else:
        print("WARNING: Could not add entry points. They were:")
        print(entry_points)
    return {}


def convert_to_flat_list(var_scripts):
    """Convert a string to a list.
    If the first element of the list is a nested list this function will
    convert it to a flat list.

    :param str/list var_scripts: Receives a string or a list to be converted
    :return list: Return a flat list
    """
    if isinstance(var_scripts, str):
        var_scripts = [var_scripts]
    elif (
        var_scripts
        and isinstance(var_scripts, list)
        and isinstance(var_scripts[0], list)
    ):
        var_scripts = [item for sublist in [s for s in var_scripts] for item in sublist]
    return var_scripts


def is_setuptools_enabled(pkginfo):
    """Function responsible to inspect if skeleton requires setuptools
    :param dict pkginfo: Dict which holds the package information
    :return Bool: Return True if it is enabled or False otherwise
    """
    entry_points = pkginfo.get("entry_points")
    if not isinstance(entry_points, dict):
        return False

    # We have *other* kinds of entry-points so we need
    # setuptools at run-time
    if set(entry_points.keys()) - {"console_scripts", "gui_scripts"}:
        return True
    return False


def valid(name):
    if re.match("[_A-Za-z][_a-zA-Z0-9]*$", name) and not keyword.iskeyword(name):
        return name
    else:
        return ""


def unpack(src_path, tempdir):
    if src_path.lower().endswith(decompressible_exts):
        tar_xf(src_path, tempdir)
    else:
        raise Exception("not a valid source: %s" % src_path)


def get_dir(tempdir):
    lst = [
        fn
        for fn in listdir(tempdir)
        if not fn.startswith(".") and isdir(join(tempdir, fn))
    ]
    if len(lst) == 1:
        dir_path = join(tempdir, lst[0])
        if isdir(dir_path):
            return dir_path
    if not lst:
        return tempdir
    raise Exception("could not find unpacked source dir")


def get_requirements(package, pkginfo, all_extras=True):
    # Look for package[extra,...] features spec:
    match_extras = re.match(r"^([^[]+)\[([^]]+)\]$", package)
    if match_extras:
        package, extras = match_extras.groups()
        extras = extras.split(",")
    else:
        extras = []

    # Extract requested extra feature requirements...
    if all_extras:
        extras_require = list(pkginfo["extras_require"].values())
    else:
        try:
            extras_require = [pkginfo["extras_require"][x] for x in extras]
        except KeyError:
            sys.exit("Error: Invalid extra features: [%s]" % ",".join(extras))
        # match PEP 508 environment markers; currently only matches the
        #  subset of environment markers that compare to python_version
        #  using a single basic Python comparison operator
        version_marker = re.compile(r"^:python_version(<|<=|!=|==|>=|>)(.+)$")
        for extra in pkginfo["extras_require"]:
            match_ver_mark = version_marker.match(extra)
            if match_ver_mark:
                op, ver = match_ver_mark.groups()
                try:
                    ver_tuple = tuple(int(x) for x in ver.strip("'\"").split("."))
                except ValueError:
                    pass  # bad match; abort
                else:
                    if op == "<":
                        satisfies_ver = sys.version_info < ver_tuple
                    elif op == "<=":
                        satisfies_ver = sys.version_info <= ver_tuple
                    elif op == "!=":
                        satisfies_ver = sys.version_info != ver_tuple
                    elif op == "==":
                        satisfies_ver = sys.version_info == ver_tuple
                    elif op == ">=":
                        satisfies_ver = sys.version_info >= ver_tuple
                    else:  # op == ">":
                        satisfies_ver = sys.version_info > ver_tuple
                    if satisfies_ver:
                        extras_require += pkginfo["extras_require"][extra]

    # ... and collect all needed requirement specs in a single list:
    requires = []
    for specs in [pkginfo.get("install_requires", "")] + extras_require:
        if isinstance(specs, str):
            requires.append(specs)
        else:
            requires.extend(specs)

    return requires


def get_pkginfo(
    package,
    filename,
    pypiurl,
    digest,
    python_version,
    extra_specs,
    config,
    setup_options,
):
    # Unfortunately, two important pieces of metadata are only stored in
    # the package itself: the dependencies, and the entry points (if the
    # package uses distribute).  Our strategy is to download the package
    # and "fake" distribute/setuptools's setup() function to get this
    # information from setup.py. If this sounds evil, keep in mind that
    # distribute itself already works by monkeypatching distutils.
    tempdir = mkdtemp("conda_skeleton_" + filename)

    if not isdir(config.src_cache):
        makedirs(config.src_cache)

    hash_type = digest[0]
    hash_value = digest[1]
    try:
        # Download it to the build source cache. That way, you have
        # it.
        download_path = join(config.src_cache, filename)
        if (
            not isfile(download_path)
            or compute_sum(download_path, hash_type) != hash_value
        ):
            download(pypiurl, join(config.src_cache, filename))
            if compute_sum(download_path, hash_type) != hash_value:
                raise RuntimeError(
                    f" Download of {package} failed"
                    f" checksum type {hash_type} expected value {hash_value}. Please"
                    " try again."
                )
        else:
            print("Using cached download")
        # Calculate the preferred hash type here if necessary.
        # Needs to be done in this block because this is where we have
        # access to the source file.
        if hash_type != "sha256":
            new_hash_value = compute_sum(download_path, "sha256")
        else:
            new_hash_value = ""

        print("Unpacking %s..." % package)
        unpack(join(config.src_cache, filename), tempdir)
        print("done")
        print("working in %s" % tempdir)
        src_dir = get_dir(tempdir)
        # TODO: find args parameters needed by run_setuppy
        run_setuppy(
            src_dir,
            tempdir,
            python_version,
            extra_specs=extra_specs,
            config=config,
            setup_options=setup_options,
        )
        try:
            with open(join(tempdir, "pkginfo.yaml")) as fn:
                pkg_info = yaml.safe_load(fn)
        except OSError:
            print(
                "WARNING: the pkginfo.yaml file was absent, falling back to pkginfo.SDist"
            )
            pkg_info = pkginfo.SDist(download_path).__dict__
        if new_hash_value:
            pkg_info["new_hash_value"] = ("sha256", new_hash_value)
    finally:
        rm_rf(tempdir)

    return pkg_info


def run_setuppy(src_dir, temp_dir, python_version, extra_specs, config, setup_options):
    """
    Patch distutils and then run setup.py in a subprocess.

    :param src_dir: Directory containing the source code
    :type src_dir: str
    :param temp_dir: Temporary directory for doing for storing pkginfo.yaml
    :type temp_dir: str
    """
    subdir = config.host_subdir
    specs = [
        f"python {python_version}*",
        "pip",
        "ruamel.yaml",
        "setuptools",
    ] + (
        ["m2-patch", "m2-gcc-libs"]
        if config.host_subdir.startswith("win")
        else ["patch"]
    )
    with open(os.path.join(src_dir, "setup.py")) as setup:
        text = setup.read()
        if "import numpy" in text or "from numpy" in text:
            specs.append("numpy")

    specs.extend(extra_specs)

    rm_rf(config.host_prefix)

    create_env(
        config.host_prefix,
        specs_or_precs=specs,
        env="host",
        subdir=subdir,
        clear_cache=False,
        config=config,
    )
    stdlib_dir = join(
        config.host_prefix,
        "Lib" if on_win else "lib/python%s" % python_version,
    )

    patch = join(temp_dir, "pypi-distutils.patch")
    with open(patch, "wb") as f:
        f.write(DISTUTILS_PATCH.format(temp_dir.replace("\\", "\\\\")).encode("utf-8"))

    # distutils deprecated in Python 3.10+, removed in Python 3.12+
    distutils = join(stdlib_dir, "distutils")
    if isdir(distutils):
        if exists(join(distutils, "core.py-copy")):
            rm_rf(join(distutils, "core.py"))
            copy2(
                join(distutils, "core.py-copy"),
                join(distutils, "core.py"),
            )
            # Avoid race conditions. Invalidate the cache.
            rm_rf(
                join(
                    distutils,
                    "__pycache__",
                    f"core.cpython-{sys.version_info[0]}{sys.version_info[1]}.pyc",
                )
            )
            rm_rf(
                join(
                    distutils,
                    "__pycache__",
                    f"core.cpython-{sys.version_info[0]}{sys.version_info[1]}.pyo",
                )
            )
        else:
            copy2(
                join(distutils, "core.py"),
                join(distutils, "core.py-copy"),
            )
        apply_patch(distutils, patch, config=config)

    setuptools = join(stdlib_dir, "site-packages", "setuptools", "_distutils")
    if isdir(setuptools):
        apply_patch(setuptools, patch, config=config)

    # Save PYTHONPATH for later
    env = os.environ.copy()
    if "PYTHONPATH" in env:
        env["PYTHONPATH"] = str(src_dir + ":" + env["PYTHONPATH"])
    else:
        env["PYTHONPATH"] = str(src_dir)
    cwd = getcwd()
    chdir(src_dir)
    cmdargs = [config.host_python, "setup.py", "install"]
    cmdargs.extend(setup_options)
    try:
        check_call_env(cmdargs, env=env)
    except subprocess.CalledProcessError:
        print("$PYTHONPATH = %s" % env["PYTHONPATH"])
        sys.exit("Error: command failed: %s" % " ".join(cmdargs))
    finally:
        chdir(cwd)


def make_entry_tests(entry_list):
    tests = []
    for entry_point in entry_list:
        entry = entry_point.partition("=")[0].strip()
        tests.append(entry + " --help")
    return tests



# Copyright (C) 2014 Anaconda, Inc
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import contextlib
import json
import logging
import multiprocessing
import os
import platform
import re
import subprocess
import sys
import warnings
from collections import defaultdict
from functools import lru_cache
from glob import glob
from logging import getLogger
from os.path import join, normpath
from typing import TYPE_CHECKING

from conda.base.constants import (
    CONDA_PACKAGE_EXTENSIONS,
    DEFAULTS_CHANNEL_NAME,
    UNKNOWN_CHANNEL,
)
from conda.common.io import env_vars
from conda.core.index import LAST_CHANNEL_URLS
from conda.core.link import PrefixSetup, UnlinkLinkTransaction
from conda.core.package_cache_data import PackageCacheData
from conda.core.prefix_data import PrefixData
from conda.exceptions import (
    CondaError,
    LinkError,
    LockError,
    NoPackagesFoundError,
    PaddingError,
    UnsatisfiableError,
)
from conda.models.channel import prioritize_channels
from conda.models.match_spec import MatchSpec

from . import utils
from .conda_interface import (
    Channel,
    PackageRecord,
    ProgressiveFetchExtract,
    TemporaryDirectory,
    context,
    create_default_packages,
    pkgs_dirs,
    reset_context,
    root_dir,
)
from .deprecations import deprecated
from .exceptions import BuildLockError, DependencyNeedsBuildingError
from .features import feature_list
from .index import get_build_index
from .os_utils import external
from .utils import (
    ensure_list,
    env_var,
    on_mac,
    on_win,
    package_record_to_requirement,
    prepend_bin_path,
)
from .variants import get_default_variant

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Iterable, TypedDict

    from .config import Config
    from .metadata import MetaData

    class InstallActionsType(TypedDict):
        PREFIX: str | os.PathLike | Path
        LINK: list[PackageRecord]


log = getLogger(__name__)

deprecated.constant("24.3", "24.5", "PREFIX_ACTION", _PREFIX_ACTION := "PREFIX")
deprecated.constant("24.3", "24.5", "LINK_ACTION", _LINK_ACTION := "LINK")

# these are things that we provide env vars for more explicitly.  This list disables the
#    pass-through of variant values to env vars for these keys.
LANGUAGES = ("PERL", "LUA", "R", "NUMPY", "PYTHON")
R_PACKAGES = ("r-base", "mro-base", "r-impl")


def get_perl_ver(config):
    return ".".join(
        config.variant.get("perl", get_default_variant(config)["perl"]).split(".")[:2]
    )


def get_lua_ver(config):
    return ".".join(
        config.variant.get("lua", get_default_variant(config)["lua"]).split(".")[:2]
    )


def get_py_ver(config):
    py = config.variant.get("python", get_default_variant(config)["python"])
    if not hasattr(py, "split"):
        py = py[0]
    return ".".join(py.split(".")[:2])


def get_r_ver(config):
    return ".".join(
        config.variant.get("r_base", get_default_variant(config)["r_base"]).split(".")[
            :3
        ]
    )


def get_npy_ver(config):
    conda_npy = "".join(
        str(config.variant.get("numpy") or get_default_variant(config)["numpy"]).split(
            "."
        )
    )
    # Convert int -> string, e.g.
    #   17 -> '1.7'
    #   110 -> '1.10'
    return conda_npy[0] + "." + conda_npy[1:]


def get_lua_include_dir(config):
    return join(config.host_prefix, "include")


@lru_cache(maxsize=None)
def verify_git_repo(
    git_exe, git_dir, git_url, git_commits_since_tag, debug=False, expected_rev="HEAD"
):
    env = os.environ.copy()
    log = utils.get_logger(__name__)

    stderr = None if debug else subprocess.DEVNULL

    if not expected_rev:
        return False

    OK = True

    env["GIT_DIR"] = git_dir
    try:
        # Verify current commit (minus our locally applied patches) matches expected commit
        current_commit = utils.check_output_env(
            [
                git_exe,
                "log",
                "-n1",
                "--format=%H",
                "HEAD" + "^" * git_commits_since_tag,
            ],
            env=env,
            stderr=stderr,
        )
        current_commit = current_commit.decode("utf-8")
        expected_tag_commit = utils.check_output_env(
            [git_exe, "log", "-n1", "--format=%H", expected_rev], env=env, stderr=stderr
        )
        expected_tag_commit = expected_tag_commit.decode("utf-8")

        if current_commit != expected_tag_commit:
            return False

        # Verify correct remote url. Need to find the git cache directory,
        # and check the remote from there.
        cache_details = utils.check_output_env(
            [git_exe, "remote", "-v"], env=env, stderr=stderr
        )
        cache_details = cache_details.decode("utf-8")
        cache_dir = cache_details.split("\n")[0].split()[1]

        if not isinstance(cache_dir, str):
            # On Windows, subprocess env can't handle unicode.
            cache_dir = cache_dir.encode(sys.getfilesystemencoding() or "utf-8")

        try:
            remote_details = utils.check_output_env(
                [git_exe, "--git-dir", cache_dir, "remote", "-v"],
                env=env,
                stderr=stderr,
            )
        except subprocess.CalledProcessError:
            if on_win and cache_dir.startswith("/"):
                cache_dir = utils.convert_unix_path_to_win(cache_dir)
            remote_details = utils.check_output_env(
                [git_exe, "--git-dir", cache_dir, "remote", "-v"],
                env=env,
                stderr=stderr,
            )
        remote_details = remote_details.decode("utf-8")
        remote_url = remote_details.split("\n")[0].split()[1]

        # on windows, remote URL comes back to us as cygwin or msys format.  Python doesn't
        # know how to normalize it.  Need to convert it to a windows path.
        if on_win and remote_url.startswith("/"):
            remote_url = utils.convert_unix_path_to_win(git_url)

        if os.path.exists(remote_url):
            # Local filepaths are allowed, but make sure we normalize them
            remote_url = normpath(remote_url)

        # If the current source directory in conda-bld/work doesn't match the user's
        # metadata git_url or git_rev, then we aren't looking at the right source.
        if not os.path.isdir(remote_url) and remote_url.lower() != git_url.lower():
            log.debug("remote does not match git_url")
            log.debug("Remote: " + remote_url.lower())
            log.debug("git_url: " + git_url.lower())
            OK = False
    except subprocess.CalledProcessError as error:
        log.debug("Error obtaining git information in verify_git_repo.  Error was: ")
        log.debug(str(error))
        OK = False
    return OK


GIT_DESCRIBE_REGEX = re.compile(
    r"(?:[_-a-zA-Z]*)"
    r"(?P<version>[a-zA-Z0-9.]+)"
    r"(?:-(?P<post>\d+)-g(?P<hash>[0-9a-f]{7,}))$"
)


def get_version_from_git_tag(tag):
    """Return a PEP440-compliant version derived from the git status.
    If that fails for any reason, return the changeset hash.
    """
    m = GIT_DESCRIBE_REGEX.match(tag)
    if m is None:
        return None
    version, post_commit, hash = m.groups()
    return version if post_commit == "0" else f"{version}.post{post_commit}+{hash}"


def get_git_info(git_exe, repo, debug):
    """
    Given a repo to a git repo, return a dictionary of:
      GIT_DESCRIBE_TAG
      GIT_DESCRIBE_TAG_PEP440
      GIT_DESCRIBE_NUMBER
      GIT_DESCRIBE_HASH
      GIT_FULL_HASH
      GIT_BUILD_STR
    from the output of git describe.
    :return:
    """
    d = {}
    log = utils.get_logger(__name__)

    stderr = None if debug else subprocess.DEVNULL

    # grab information from describe
    env = os.environ.copy()
    env["GIT_DIR"] = repo
    keys = ["GIT_DESCRIBE_TAG", "GIT_DESCRIBE_NUMBER", "GIT_DESCRIBE_HASH"]

    try:
        output = utils.check_output_env(
            [git_exe, "describe", "--tags", "--long", "HEAD"],
            env=env,
            cwd=os.path.dirname(repo),
            stderr=stderr,
        ).splitlines()[0]
        output = output.decode("utf-8")
        parts = output.rsplit("-", 2)
        if len(parts) == 3:
            d.update(dict(zip(keys, parts)))
        d["GIT_DESCRIBE_TAG_PEP440"] = str(get_version_from_git_tag(output))
    except subprocess.CalledProcessError:
        msg = (
            "Failed to obtain git tag information.\n"
            "Consider using annotated tags if you are not already "
            "as they are more reliable when used with git describe."
        )
        log.debug(msg)

    # If there was no tag reachable from HEAD, the above failed and the short hash is not set.
    # Try to get the short hash from describing with all refs (not just the tags).
    if "GIT_DESCRIBE_HASH" not in d:
        try:
            output = utils.check_output_env(
                [git_exe, "describe", "--all", "--long", "HEAD"],
                env=env,
                cwd=os.path.dirname(repo),
                stderr=stderr,
            ).splitlines()[0]
            output = output.decode("utf-8")
            parts = output.rsplit("-", 2)
            if len(parts) == 3:
                # Don't save GIT_DESCRIBE_TAG and GIT_DESCRIBE_NUMBER because git (probably)
                # described a branch. We just want to save the short hash.
                d["GIT_DESCRIBE_HASH"] = parts[-1]
        except subprocess.CalledProcessError as error:
            log.debug("Error obtaining git commit information.  Error was: ")
            log.debug(str(error))

    try:
        # get the _full_ hash of the current HEAD
        output = utils.check_output_env(
            [git_exe, "rev-parse", "HEAD"],
            env=env,
            cwd=os.path.dirname(repo),
            stderr=stderr,
        ).splitlines()[0]
        output = output.decode("utf-8")

        d["GIT_FULL_HASH"] = output
    except subprocess.CalledProcessError as error:
        log.debug("Error obtaining git commit information.  Error was: ")
        log.debug(str(error))

    # set up the build string
    if "GIT_DESCRIBE_NUMBER" in d and "GIT_DESCRIBE_HASH" in d:
        d["GIT_BUILD_STR"] = "{}_{}".format(
            d["GIT_DESCRIBE_NUMBER"], d["GIT_DESCRIBE_HASH"]
        )

    # issues on Windows with the next line of the command prompt being recorded here.
    assert not any("\n" in value for value in d.values())
    return d


def get_hg_build_info(repo):
    env = os.environ.copy()
    env["HG_DIR"] = repo
    env = {str(key): str(value) for key, value in env.items()}

    d = {}
    cmd = [
        "hg",
        "log",
        "--template",
        "{rev}|{node|short}|{latesttag}|{latesttagdistance}|{branch}",
        "--rev",
        ".",
    ]
    output = utils.check_output_env(cmd, env=env, cwd=os.path.dirname(repo))
    output = output.decode("utf-8")
    rev, short_id, tag, distance, branch = output.split("|")
    if tag != "null":
        d["HG_LATEST_TAG"] = tag
    if branch == "":
        branch = "default"
    d["HG_BRANCH"] = branch
    d["HG_NUM_ID"] = rev
    d["HG_LATEST_TAG_DISTANCE"] = distance
    d["HG_SHORT_ID"] = short_id
    d["HG_BUILD_STR"] = "{}_{}".format(d["HG_NUM_ID"], d["HG_SHORT_ID"])
    return d


def get_dict(
    m,
    prefix=None,
    for_env=True,
    skip_build_id=False,
    escape_backslash=False,
    variant=None,
):
    if not prefix:
        prefix = m.config.host_prefix

    m.config._merge_build_host = m.build_is_host

    # conda-build specific vars
    d = conda_build_vars(prefix, m.config)

    # languages
    d.update(python_vars(m, prefix, escape_backslash))
    d.update(perl_vars(m, prefix, escape_backslash))
    d.update(lua_vars(m, prefix, escape_backslash))
    d.update(r_vars(m, prefix, escape_backslash))

    if m:
        d.update(meta_vars(m, skip_build_id=skip_build_id))

    # system
    d.update(os_vars(m, prefix))

    # features
    d.update({feat.upper(): str(int(value)) for feat, value in feature_list})

    variant = variant or m.config.variant
    for k, v in variant.items():
        if not for_env or (k.upper() not in d and k.upper() not in LANGUAGES):
            d[k] = v
    return d


def conda_build_vars(prefix, config):
    src_dir = (
        config.test_dir if os.path.basename(prefix)[:2] == "_t" else config.work_dir
    )
    return {
        "CONDA_BUILD": "1",
        "PYTHONNOUSERSITE": "1",
        "CONDA_DEFAULT_ENV": config.host_prefix,
        "ARCH": str(config.host_arch),
        # This is the one that is most important for where people put artifacts that get bundled.
        #     It is fed from our function argument, and can be any of:
        #     1. Build prefix - when host requirements are not explicitly set,
        #        then prefix = build prefix = host prefix
        #     2. Host prefix - when host requirements are explicitly set, prefix = host prefix
        #     3. Test prefix - during test runs, this points at the test prefix
        "PREFIX": prefix,
        # This is for things that are specifically build tools.  Things that run on the build
        #    platform, but probably should not be linked against, since they may not run on the
        #    destination host platform
        # It can be equivalent to config.host_prefix if the host section is not explicitly set.
        "BUILD_PREFIX": config.build_prefix,
        "SYS_PREFIX": sys.prefix,
        "SYS_PYTHON": sys.executable,
        "SUBDIR": config.host_subdir,
        "build_platform": config.build_subdir,
        "SRC_DIR": src_dir,
        "HTTPS_PROXY": os.getenv("HTTPS_PROXY", ""),
        "HTTP_PROXY": os.getenv("HTTP_PROXY", ""),
        "REQUESTS_CA_BUNDLE": os.getenv("REQUESTS_CA_BUNDLE", ""),
        "DIRTY": "1" if config.dirty else "",
        "ROOT": root_dir,
    }


def python_vars(metadata, prefix, escape_backslash):
    py_ver = get_py_ver(metadata.config)
    stdlib_dir = utils.get_stdlib_dir(prefix, py_ver)
    sp_dir = utils.get_site_packages(prefix, py_ver)

    if utils.on_win and escape_backslash:
        stdlib_dir = stdlib_dir.replace("\\", "\\\\")
        sp_dir = sp_dir.replace("\\", "\\\\")

    vars_ = {
        "CONDA_PY": "".join(py_ver.split(".")[:2]),
        "PY3K": str(int(int(py_ver[0]) >= 3)),
        "PY_VER": py_ver,
        "STDLIB_DIR": stdlib_dir,
        "SP_DIR": sp_dir,
    }
    build_or_host = "host" if metadata.is_cross else "build"
    deps = [str(ms.name) for ms in metadata.ms_depends(build_or_host)]
    if "python" in deps or metadata.name() == "python":
        python_bin = metadata.config.python_bin(prefix, metadata.config.host_subdir)

        if utils.on_win and escape_backslash:
            python_bin = python_bin.replace("\\", "\\\\")

        vars_.update(
            {
                # host prefix is always fine, because it is the same as build when is_cross is False
                "PYTHON": python_bin,
            }
        )

    np_ver = metadata.config.variant.get(
        "numpy", get_default_variant(metadata.config)["numpy"]
    )
    vars_["NPY_VER"] = ".".join(np_ver.split(".")[:2])
    vars_["CONDA_NPY"] = "".join(np_ver.split(".")[:2])
    vars_["NPY_DISTUTILS_APPEND_FLAGS"] = "1"
    return vars_


def perl_vars(metadata, prefix, escape_backslash):
    vars_ = {
        "PERL_VER": get_perl_ver(metadata.config),
        "CONDA_PERL": get_perl_ver(metadata.config),
    }
    build_or_host = "host" if metadata.is_cross else "build"
    deps = [str(ms.name) for ms in metadata.ms_depends(build_or_host)]
    if "perl" in deps or metadata.name() == "perl":
        perl_bin = metadata.config.perl_bin(prefix, metadata.config.host_subdir)

        if utils.on_win and escape_backslash:
            perl_bin = perl_bin.replace("\\", "\\\\")

        vars_.update(
            {
                # host prefix is always fine, because it is the same as build when is_cross is False
                "PERL": perl_bin,
            }
        )
    return vars_


def lua_vars(metadata, prefix, escape_backslash):
    vars_ = {
        "LUA_VER": get_lua_ver(metadata.config),
        "CONDA_LUA": get_lua_ver(metadata.config),
    }
    build_or_host = "host" if metadata.is_cross else "build"
    deps = [str(ms.name) for ms in metadata.ms_depends(build_or_host)]
    if "lua" in deps:
        lua_bin = metadata.config.lua_bin(prefix, metadata.config.host_subdir)
        lua_include_dir = get_lua_include_dir(metadata.config)

        if utils.on_win and escape_backslash:
            lua_bin = lua_bin.replace("\\", "\\\\")
            lua_include_dir = lua_include_dir.replace("\\", "\\\\")

        vars_.update(
            {
                "LUA": lua_bin,
                "LUA_INCLUDE_DIR": lua_include_dir,
            }
        )
    return vars_


def r_vars(metadata, prefix, escape_backslash):
    vars_ = {
        "R_VER": get_r_ver(metadata.config),
        "CONDA_R": get_r_ver(metadata.config),
    }

    build_or_host = "host" if metadata.is_cross else "build"
    deps = [str(ms.name) for ms in metadata.ms_depends(build_or_host)]
    if any(r_pkg in deps for r_pkg in R_PACKAGES) or metadata.name() in R_PACKAGES:
        r_bin = metadata.config.r_bin(prefix, metadata.config.host_subdir)
        # set R_USER explicitly to prevent crosstalk with existing R_LIBS_USER packages
        r_user = join(prefix, "Libs", "R")

        if utils.on_win and escape_backslash:
            r_bin = r_bin.replace("\\", "\\\\")

        vars_.update(
            {
                "R": r_bin,
                "R_USER": r_user,
            }
        )
    return vars_


def meta_vars(meta: MetaData, skip_build_id=False):
    d = {}
    for var_name in ensure_list(meta.get_value("build/script_env", [])):
        if "=" in var_name:
            var_name, value = var_name.split("=", 1)
        else:
            value = os.getenv(var_name)
        if value is None:
            warnings.warn(
                "The environment variable '%s' specified in script_env is undefined."
                % var_name,
                UserWarning,
            )
        else:
            d[var_name] = value
            warnings.warn(
                f"The environment variable '{var_name}' is being passed through with value "
                f"'{'<hidden>' if meta.config.suppress_variables else value}'.  "
                "If you are splitting build and test phases with --no-test, please ensure "
                "that this value is also set similarly at test time.",
                UserWarning,
            )

    folder = meta.get_value("source/0/folder", "")
    repo_dir = join(meta.config.work_dir, folder)
    git_dir = join(repo_dir, ".git")
    hg_dir = join(repo_dir, ".hg")

    if not isinstance(git_dir, str):
        # On Windows, subprocess env can't handle unicode.
        git_dir = git_dir.encode(sys.getfilesystemencoding() or "utf-8")

    git_exe = external.find_executable("git", meta.config.build_prefix)
    if git_exe and os.path.exists(git_dir):
        # We set all 'source' metavars using the FIRST source entry in meta.yaml.
        git_url = meta.get_value("source/0/git_url")

        if os.path.exists(git_url):
            if on_win:
                git_url = utils.convert_unix_path_to_win(git_url)
            # If git_url is a relative path instead of a url, convert it to an abspath
            git_url = normpath(join(meta.path, git_url))

        _x = False

        if git_url:
            _x = verify_git_repo(
                git_exe,
                git_dir,
                git_url,
                meta.config.git_commits_since_tag,
                meta.config.debug,
                meta.get_value("source/0/git_rev", "HEAD"),
            )

        if _x or meta.get_value("source/0/path"):
            d.update(get_git_info(git_exe, git_dir, meta.config.debug))

    elif external.find_executable("hg", meta.config.build_prefix) and os.path.exists(
        hg_dir
    ):
        d.update(get_hg_build_info(hg_dir))

    d["PKG_NAME"] = meta.name()
    d["PKG_VERSION"] = meta.version()
    d["PKG_BUILDNUM"] = str(meta.build_number())
    if meta.final and not skip_build_id:
        d["PKG_BUILD_STRING"] = meta.build_id()
        d["PKG_HASH"] = meta.hash_dependencies()
    else:
        d["PKG_BUILD_STRING"] = "placeholder"
        d["PKG_HASH"] = "1234567"
    d["RECIPE_DIR"] = meta.path
    return d


@lru_cache(maxsize=None)
def get_cpu_count():
    if on_mac:
        # multiprocessing.cpu_count() is not reliable on OSX
        # See issue #645 on github.com/conda/conda-build
        out, _ = subprocess.Popen(
            "sysctl -n hw.logicalcpu", shell=True, stdout=subprocess.PIPE
        ).communicate()
        return out.decode("utf-8").strip()
    else:
        try:
            return str(multiprocessing.cpu_count())
        except NotImplementedError:
            return "1"


def get_shlib_ext(host_platform):
    # Return the shared library extension.
    if host_platform.startswith("win"):
        return ".dll"
    elif host_platform in ["osx", "darwin"]:
        return ".dylib"
    elif host_platform.startswith("linux") or host_platform.endswith("-wasm32"):
        return ".so"
    elif host_platform == "noarch":
        # noarch packages should not contain shared libraries, use the system
        # platform if this is requested
        return get_shlib_ext(sys.platform)
    else:
        raise NotImplementedError(host_platform)


def windows_vars(m, get_default, prefix):
    """This is setting variables on a dict that is part of the get_default function"""
    # We have gone for the clang values here.
    win_arch = "i386" if str(m.config.host_arch) == "32" else "amd64"
    win_msvc = "19.0.0"
    library_prefix = join(prefix, "Library")
    drive, tail = m.config.host_prefix.split(":")
    get_default("SCRIPTS", join(prefix, "Scripts"))
    get_default("LIBRARY_PREFIX", library_prefix)
    get_default("LIBRARY_BIN", join(library_prefix, "bin"))
    get_default("LIBRARY_INC", join(library_prefix, "include"))
    get_default("LIBRARY_LIB", join(library_prefix, "lib"))
    get_default(
        "CYGWIN_PREFIX", "".join(("/cygdrive/", drive.lower(), tail.replace("\\", "/")))
    )
    # see https://en.wikipedia.org/wiki/Environment_variable#Default_values
    get_default("ALLUSERSPROFILE")
    get_default("APPDATA")
    get_default("CommonProgramFiles")
    get_default("CommonProgramFiles(x86)")
    get_default("CommonProgramW6432")
    get_default("COMPUTERNAME")
    get_default("ComSpec")
    get_default("HOMEDRIVE")
    get_default("HOMEPATH")
    get_default("LOCALAPPDATA")
    get_default("LOGONSERVER")
    get_default("NUMBER_OF_PROCESSORS")
    get_default("PATHEXT")
    get_default("ProgramData")
    get_default("ProgramFiles")
    get_default("ProgramFiles(x86)")
    get_default("ProgramW6432")
    get_default("PROMPT")
    get_default("PSModulePath")
    get_default("PUBLIC")
    get_default("SystemDrive")
    get_default("SystemRoot")
    get_default("TEMP")
    get_default("TMP")
    get_default("USERDOMAIN")
    get_default("USERNAME")
    get_default("USERPROFILE")
    get_default("windir")
    # CPU data, see https://github.com/conda/conda-build/issues/2064
    get_default("PROCESSOR_ARCHITEW6432")
    get_default("PROCESSOR_ARCHITECTURE")
    get_default("PROCESSOR_IDENTIFIER")
    get_default("BUILD", win_arch + "-pc-windows-" + win_msvc)
    for k in os.environ.keys():
        if re.match("VS[0-9]{2,3}COMNTOOLS", k):
            get_default(k)
        elif re.match("VS[0-9]{4}INSTALLDIR", k):
            get_default(k)


def unix_vars(m, get_default, prefix):
    """This is setting variables on a dict that is part of the get_default function"""
    get_default("HOME", "UNKNOWN")
    get_default("PKG_CONFIG_PATH", join(prefix, "lib", "pkgconfig"))
    get_default("CMAKE_GENERATOR", "Unix Makefiles")
    get_default("SSL_CERT_FILE")


def osx_vars(m, get_default, prefix):
    """This is setting variables on a dict that is part of the get_default function"""
    if str(m.config.host_arch) == "32":
        OSX_ARCH = "i386"
        MACOSX_DEPLOYMENT_TARGET = 10.9
    elif str(m.config.host_arch) == "arm64":
        OSX_ARCH = "arm64"
        MACOSX_DEPLOYMENT_TARGET = 11.0
    else:
        OSX_ARCH = "x86_64"
        MACOSX_DEPLOYMENT_TARGET = 10.9

    if str(m.config.arch) == "32":
        BUILD = "i386-apple-darwin13.4.0"
    elif str(m.config.arch) == "arm64":
        BUILD = "arm64-apple-darwin20.0.0"
    else:
        BUILD = "x86_64-apple-darwin13.4.0"
    # 10.7 install_name_tool -delete_rpath causes broken dylibs, I will revisit this ASAP.
    # rpath = ' -Wl,-rpath,%(PREFIX)s/lib' % d # SIP workaround, DYLD_* no longer works.
    # d['LDFLAGS'] = ldflags + rpath + ' -arch %(OSX_ARCH)s' % d
    get_default("OSX_ARCH", OSX_ARCH)
    get_default("MACOSX_DEPLOYMENT_TARGET", MACOSX_DEPLOYMENT_TARGET)
    get_default("BUILD", BUILD)


@lru_cache(maxsize=None)
def _machine_and_architecture():
    return platform.machine(), platform.architecture()


def linux_vars(m, get_default, prefix):
    """This is setting variables on a dict that is part of the get_default function"""
    platform_machine, platform_architecture = _machine_and_architecture()
    build_arch = platform_machine
    # Python reports x86_64 when running a i686 Python binary on a 64-bit CPU
    # unless run through linux32. Issue a warning when we detect this.
    if build_arch == "x86_64" and platform_architecture[0] == "32bit":
        print("Warning: You are running 32-bit Python on a 64-bit linux installation")
        print("         but have not launched it via linux32. Various qeuries *will*")
        print("         give unexpected results (uname -m, platform.machine() etc)")
        build_arch = "i686"
    # the GNU triplet is powerpc, not ppc. This matters.
    if build_arch.startswith("ppc"):
        build_arch = build_arch.replace("ppc", "powerpc")
    if (
        build_arch.startswith("powerpc")
        or build_arch.startswith("aarch64")
        or build_arch.startswith("s390x")
    ):
        build_distro = "cos7"
    else:
        build_distro = "cos6"
    # There is also QEMU_SET_ENV, but that needs to be
    # filtered so it only contains the result of `linux_vars`
    # which, before this change was empty, and after it only
    # contains other QEMU env vars.
    get_default("CFLAGS")
    get_default("CXXFLAGS")
    get_default("LDFLAGS")
    get_default("QEMU_LD_PREFIX")
    get_default("QEMU_UNAME")
    get_default("DEJAGNU")
    get_default("DISPLAY")
    get_default("LD_RUN_PATH", prefix + "/lib")
    get_default("BUILD", build_arch + "-conda_" + build_distro + "-linux-gnu")


def set_from_os_or_variant(out_dict, key, variant, default):
    value = os.getenv(key)
    if not value:
        value = variant.get(key, default)
    if value:
        out_dict[key] = value


def system_vars(env_dict, m, prefix):
    warnings.warn(
        "`conda_build.environ.system_vars` is pending deprecation and will be removed in a "
        "future release. Please use `conda_build.environ.os_vars` instead.",
        PendingDeprecationWarning,
    )
    return os_vars(m, prefix)


def os_vars(m, prefix):
    d = dict()
    # note the dictionary is passed in here - variables are set in that dict if they are non-null
    get_default = lambda key, default="": set_from_os_or_variant(
        d, key, m.config.variant, default
    )

    get_default("CPU_COUNT", get_cpu_count())
    get_default("LANG")
    get_default("LC_ALL")
    get_default("MAKEFLAGS")
    d["SHLIB_EXT"] = get_shlib_ext(m.config.host_platform)
    d["PATH"] = os.environ.copy()["PATH"]

    if not m.config.activate:
        d = prepend_bin_path(d, m.config.host_prefix)

    if on_win:
        windows_vars(m, get_default, prefix)
    else:
        unix_vars(m, get_default, prefix)

    if m.config.host_platform == "osx":
        osx_vars(m, get_default, prefix)
    elif m.config.host_platform == "linux":
        linux_vars(m, get_default, prefix)

    return d


@deprecated("24.3", "24.5")
class InvalidEnvironment(Exception):
    pass


# Stripped-down Environment class from conda-tools ( https://github.com/groutr/conda-tools )
# Vendored here to avoid the whole dependency for just this bit.
@deprecated("24.3", "24.5")
def _load_json(path):
    with open(path) as fin:
        x = json.load(fin)
    return x


@deprecated("24.3", "24.5")
def _load_all_json(path):
    """
    Load all json files in a directory.  Return dictionary with filenames mapped to json
    dictionaries.
    """
    root, _, files = next(utils.walk(path))
    result = {}
    for f in files:
        if f.endswith(".json"):
            result[f] = _load_json(join(root, f))
    return result


@deprecated("24.3", "24.5", addendum="Use `conda.core.prefix_data.PrefixData` instead.")
class Environment:
    def __init__(self, path):
        """
        Initialize an Environment object.

        To reflect changes in the underlying environment, a new Environment object should be
        created.
        """
        self.path = path
        self._meta = join(path, "conda-meta")
        if os.path.isdir(path) and os.path.isdir(self._meta):
            self._packages = {}
        else:
            raise InvalidEnvironment(f"Unable to load environment {path}")

    def _read_package_json(self):
        if not self._packages:
            self._packages = _load_all_json(self._meta)

    def package_specs(self):
        """
        List all package specs in the environment.
        """
        self._read_package_json()
        json_objs = self._packages.values()
        specs = []
        for i in json_objs:
            p, v, b = i["name"], i["version"], i["build"]
            specs.append(f"{p} {v} {b}")
        return specs


cached_precs: dict[
    tuple[tuple[str | MatchSpec, ...], Any, Any, Any, bool], list[PackageRecord]
] = {}
deprecated.constant("24.3", "24.5", "cached_actions", cached_precs)
last_index_ts = 0


# NOTE: The function has to retain the "get_install_actions" name for now since
#       conda_libmamba_solver.solver.LibMambaSolver._called_from_conda_build
#       checks for this name in the call stack explicitly.
def get_install_actions(
    prefix: str | os.PathLike | Path,
    specs: Iterable[str | MatchSpec],
    env,  # unused
    retries: int = 0,
    subdir=None,
    verbose: bool = True,
    debug: bool = False,
    locking: bool = True,
    bldpkgs_dirs=None,
    timeout=900,
    disable_pip: bool = False,
    max_env_retry: int = 3,
    output_folder=None,
    channel_urls=None,
) -> list[PackageRecord]:
    global cached_precs
    global last_index_ts

    log = utils.get_logger(__name__)
    conda_log_level = logging.WARN
    specs = list(specs)
    if specs:
        specs.extend(create_default_packages)
    if verbose or debug:
        capture = contextlib.nullcontext
        if debug:
            conda_log_level = logging.DEBUG
    else:
        capture = utils.capture
    for feature, value in feature_list:
        if value:
            specs.append("%s@" % feature)

    bldpkgs_dirs = ensure_list(bldpkgs_dirs)

    index, index_ts, _ = get_build_index(
        subdir,
        list(bldpkgs_dirs)[0],
        output_folder=output_folder,
        channel_urls=channel_urls,
        debug=debug,
        verbose=verbose,
        locking=locking,
        timeout=timeout,
    )
    specs = tuple(
        utils.ensure_valid_spec(spec) for spec in specs if not str(spec).endswith("@")
    )

    precs: list[PackageRecord] = []
    if (
        specs,
        env,
        subdir,
        channel_urls,
        disable_pip,
    ) in cached_precs and last_index_ts >= index_ts:
        precs = cached_precs[(specs, env, subdir, channel_urls, disable_pip)].copy()
    elif specs:
        # this is hiding output like:
        #    Fetching package metadata ...........
        #    Solving package specifications: ..........
        with utils.LoggingContext(conda_log_level):
            with capture():
                try:
                    precs = _install_actions(prefix, index, specs)["LINK"]
                except (NoPackagesFoundError, UnsatisfiableError) as exc:
                    raise DependencyNeedsBuildingError(exc, subdir=subdir)
                except (
                    SystemExit,
                    PaddingError,
                    LinkError,
                    DependencyNeedsBuildingError,
                    CondaError,
                    AssertionError,
                    BuildLockError,
                ) as exc:
                    if "lock" in str(exc):
                        log.warn(
                            "failed to get package records, retrying.  exception was: %s",
                            str(exc),
                        )
                    elif (
                        "requires a minimum conda version" in str(exc)
                        or "link a source that does not" in str(exc)
                        or isinstance(exc, AssertionError)
                    ):
                        locks = utils.get_conda_operation_locks(
                            locking, bldpkgs_dirs, timeout
                        )
                        with utils.try_acquire_locks(locks, timeout=timeout):
                            pkg_dir = str(exc)
                            folder = 0
                            while (
                                os.path.dirname(pkg_dir) not in pkgs_dirs
                                and folder < 20
                            ):
                                pkg_dir = os.path.dirname(pkg_dir)
                                folder += 1
                            log.warn(
                                "I think conda ended up with a partial extraction for %s. "
                                "Removing the folder and retrying",
                                pkg_dir,
                            )
                            if pkg_dir in pkgs_dirs and os.path.isdir(pkg_dir):
                                utils.rm_rf(pkg_dir)
                    if retries < max_env_retry:
                        log.warn(
                            "failed to get package records, retrying.  exception was: %s",
                            str(exc),
                        )
                        precs = get_package_records(
                            prefix,
                            specs,
                            env,
                            retries=retries + 1,
                            subdir=subdir,
                            verbose=verbose,
                            debug=debug,
                            locking=locking,
                            bldpkgs_dirs=tuple(bldpkgs_dirs),
                            timeout=timeout,
                            disable_pip=disable_pip,
                            max_env_retry=max_env_retry,
                            output_folder=output_folder,
                            channel_urls=tuple(channel_urls),
                        )
                    else:
                        log.error(
                            "Failed to get package records, max retries exceeded."
                        )
                        raise
        if disable_pip:
            for pkg in ("pip", "setuptools", "wheel"):
                # specs are the raw specifications, not the conda-derived actual specs
                #   We're testing that pip etc. are manually specified
                if not any(
                    re.match(r"^%s(?:$|[\s=].*)" % pkg, str(dep)) for dep in specs
                ):
                    precs = [prec for prec in precs if prec.name != pkg]
        cached_precs[(specs, env, subdir, channel_urls, disable_pip)] = precs.copy()
        last_index_ts = index_ts
    return precs


get_package_records = get_install_actions
del get_install_actions


def create_env(
    prefix: str | os.PathLike | Path,
    specs_or_precs: Iterable[str | MatchSpec] | Iterable[PackageRecord],
    env,
    config,
    subdir,
    clear_cache: bool = True,
    retry: int = 0,
    locks=None,
    is_cross: bool = False,
    is_conda: bool = False,
) -> None:
    """
    Create a conda envrionment for the given prefix and specs.
    """
    if config.debug:
        external_logger_context = utils.LoggingContext(logging.DEBUG)
    else:
        external_logger_context = utils.LoggingContext(logging.WARN)

    if os.path.exists(prefix):
        for entry in glob(os.path.join(prefix, "*")):
            utils.rm_rf(entry)

    with external_logger_context:
        log = utils.get_logger(__name__)

        # if os.path.isdir(prefix):
        #     utils.rm_rf(prefix)

        specs_or_precs = tuple(ensure_list(specs_or_precs))
        if specs_or_precs:  # Don't waste time if there is nothing to do
            log.debug("Creating environment in %s", prefix)
            log.debug(str(specs_or_precs))

            if not locks:
                locks = utils.get_conda_operation_locks(config)
            try:
                with utils.try_acquire_locks(locks, timeout=config.timeout):
                    # input is a list of specs in MatchSpec format
                    if not isinstance(specs_or_precs[0], PackageRecord):
                        precs = get_package_records(
                            prefix,
                            tuple(set(specs_or_precs)),
                            env,
                            subdir=subdir,
                            verbose=config.verbose,
                            debug=config.debug,
                            locking=config.locking,
                            bldpkgs_dirs=tuple(config.bldpkgs_dirs),
                            timeout=config.timeout,
                            disable_pip=config.disable_pip,
                            max_env_retry=config.max_env_retry,
                            output_folder=config.output_folder,
                            channel_urls=tuple(config.channel_urls),
                        )
                    else:
                        precs = specs_or_precs
                    index, _, _ = get_build_index(
                        subdir=subdir,
                        bldpkgs_dir=config.bldpkgs_dir,
                        output_folder=config.output_folder,
                        channel_urls=config.channel_urls,
                        debug=config.debug,
                        verbose=config.verbose,
                        locking=config.locking,
                        timeout=config.timeout,
                    )
                    _display_actions(prefix, precs)
                    if utils.on_win:
                        for k, v in os.environ.items():
                            os.environ[k] = str(v)
                    with env_var("CONDA_QUIET", not config.verbose, reset_context):
                        with env_var("CONDA_JSON", not config.verbose, reset_context):
                            _execute_actions(prefix, precs)
            except (
                SystemExit,
                PaddingError,
                LinkError,
                DependencyNeedsBuildingError,
                CondaError,
                BuildLockError,
            ) as exc:
                if (
                    "too short in" in str(exc)
                    or re.search(
                        "post-link failed for: (?:[a-zA-Z]*::)?openssl", str(exc)
                    )
                    or isinstance(exc, PaddingError)
                ) and config.prefix_length > 80:
                    if config.prefix_length_fallback:
                        log.warn(
                            "Build prefix failed with prefix length %d",
                            config.prefix_length,
                        )
                        log.warn("Error was: ")
                        log.warn(str(exc))
                        log.warn(
                            "One or more of your package dependencies needs to be rebuilt "
                            "with a longer prefix length."
                        )
                        log.warn(
                            "Falling back to legacy prefix length of 80 characters."
                        )
                        log.warn(
                            "Your package will not install into prefixes > 80 characters."
                        )
                        config.prefix_length = 80

                        create_env(
                            (
                                config.host_prefix
                                if "_h_env" in prefix
                                else config.build_prefix
                            ),
                            specs_or_precs,
                            config=config,
                            subdir=subdir,
                            env=env,
                            clear_cache=clear_cache,
                            is_cross=is_cross,
                        )
                    else:
                        raise
                elif "lock" in str(exc):
                    if retry < config.max_env_retry:
                        log.warn(
                            "failed to create env, retrying.  exception was: %s",
                            str(exc),
                        )
                        create_env(
                            prefix,
                            specs_or_precs,
                            config=config,
                            subdir=subdir,
                            env=env,
                            clear_cache=clear_cache,
                            retry=retry + 1,
                            is_cross=is_cross,
                        )
                elif "requires a minimum conda version" in str(
                    exc
                ) or "link a source that does not" in str(exc):
                    with utils.try_acquire_locks(locks, timeout=config.timeout):
                        pkg_dir = str(exc)
                        folder = 0
                        while os.path.dirname(pkg_dir) not in pkgs_dirs and folder < 20:
                            pkg_dir = os.path.dirname(pkg_dir)
                            folder += 1
                        log.warn(
                            "I think conda ended up with a partial extraction for %s.  "
                            "Removing the folder and retrying",
                            pkg_dir,
                        )
                        if os.path.isdir(pkg_dir):
                            utils.rm_rf(pkg_dir)
                    if retry < config.max_env_retry:
                        log.warn(
                            "failed to create env, retrying.  exception was: %s",
                            str(exc),
                        )
                        create_env(
                            prefix,
                            specs_or_precs,
                            config=config,
                            subdir=subdir,
                            env=env,
                            clear_cache=clear_cache,
                            retry=retry + 1,
                            is_cross=is_cross,
                        )
                    else:
                        log.error("Failed to create env, max retries exceeded.")
                        raise
                else:
                    raise
            # HACK: some of the time, conda screws up somehow and incomplete packages result.
            #    Just retry.
            except (
                AssertionError,
                OSError,
                ValueError,
                RuntimeError,
                LockError,
            ) as exc:
                if isinstance(exc, AssertionError):
                    with utils.try_acquire_locks(locks, timeout=config.timeout):
                        pkg_dir = os.path.dirname(os.path.dirname(str(exc)))
                        log.warn(
                            "I think conda ended up with a partial extraction for %s.  "
                            "Removing the folder and retrying",
                            pkg_dir,
                        )
                        if os.path.isdir(pkg_dir):
                            utils.rm_rf(pkg_dir)
                if retry < config.max_env_retry:
                    log.warn(
                        "failed to create env, retrying.  exception was: %s", str(exc)
                    )
                    create_env(
                        prefix,
                        specs_or_precs,
                        config=config,
                        subdir=subdir,
                        env=env,
                        clear_cache=clear_cache,
                        retry=retry + 1,
                        is_cross=is_cross,
                    )
                else:
                    log.error("Failed to create env, max retries exceeded.")
                    raise


def get_pkg_dirs_locks(dirs, config):
    return [utils.get_lock(folder, timeout=config.timeout) for folder in dirs]


def clean_pkg_cache(dist: str, config: Config) -> None:
    with utils.LoggingContext(logging.DEBUG if config.debug else logging.WARN):
        locks = get_pkg_dirs_locks([config.bldpkgs_dir] + pkgs_dirs, config)
        with utils.try_acquire_locks(locks, timeout=config.timeout):
            for pkgs_dir in pkgs_dirs:
                if any(
                    os.path.exists(os.path.join(pkgs_dir, f"{dist}{ext}"))
                    for ext in ("", *CONDA_PACKAGE_EXTENSIONS)
                ):
                    log.debug(
                        "Conda caching error: %s package remains in cache after removal",
                        dist,
                    )
                    log.debug("manually removing to compensate")
                    package_cache = PackageCacheData.first_writable([pkgs_dir])
                    for cache_pkg_id in package_cache.query(dist):
                        package_cache.remove(cache_pkg_id)

        # Note that this call acquires the relevant locks, so this must be called
        # outside the lock context above.
        remove_existing_packages(pkgs_dirs, [dist], config)


def remove_existing_packages(dirs, fns, config):
    locks = get_pkg_dirs_locks(dirs, config) if config.locking else []

    with utils.try_acquire_locks(locks, timeout=config.timeout):
        for folder in dirs:
            for fn in fns:
                all_files = [fn]
                if not os.path.isabs(fn):
                    all_files = glob(os.path.join(folder, fn + "*"))
                for entry in all_files:
                    utils.rm_rf(entry)


def get_pinned_deps(m, section):
    with TemporaryDirectory(prefix="_") as tmpdir:
        precs = get_package_records(
            tmpdir,
            tuple(m.ms_depends(section)),
            section,
            subdir=m.config.target_subdir,
            debug=m.config.debug,
            verbose=m.config.verbose,
            locking=m.config.locking,
            bldpkgs_dirs=tuple(m.config.bldpkgs_dirs),
            timeout=m.config.timeout,
            disable_pip=m.config.disable_pip,
            max_env_retry=m.config.max_env_retry,
            output_folder=m.config.output_folder,
            channel_urls=tuple(m.config.channel_urls),
        )
    return [package_record_to_requirement(prec) for prec in precs]


# NOTE: The function has to retain the "install_actions" name for now since
#       conda_libmamba_solver.solver.LibMambaSolver._called_from_conda_build
#       checks for this name in the call stack explicitly.
def install_actions(
    prefix: str | os.PathLike | Path,
    index,
    specs: Iterable[str | MatchSpec],
) -> InstallActionsType:
    # This is copied over from https://github.com/conda/conda/blob/23.11.0/conda/plan.py#L471
    # but reduced to only the functionality actually used within conda-build.

    with env_vars(
        {
            "CONDA_ALLOW_NON_CHANNEL_URLS": "true",
            "CONDA_SOLVER_IGNORE_TIMESTAMPS": "false",
        },
        callback=reset_context,
    ):
        # a hack since in conda-build we don't track channel_priority_map
        channels: tuple[Channel, ...] | None
        subdirs: tuple[str, ...] | None
        if LAST_CHANNEL_URLS:
            channel_priority_map = prioritize_channels(LAST_CHANNEL_URLS)
            # tuple(dict.fromkeys(...)) removes duplicates while preserving input order.
            channels = tuple(
                dict.fromkeys(Channel(url) for url in channel_priority_map)
            )
            subdirs = (
                tuple(
                    dict.fromkeys(
                        subdir for channel in channels if (subdir := channel.subdir)
                    )
                )
                or context.subdirs
            )
        else:
            channels = subdirs = None

        mspecs = tuple(MatchSpec(spec) for spec in specs)

        PrefixData._cache_.clear()

        solver_backend = context.plugin_manager.get_cached_solver_backend()
        solver = solver_backend(prefix, channels, subdirs, specs_to_add=mspecs)
        if index:
            # Solver can modify the index (e.g., Solver._prepare adds virtual
            # package) => Copy index (just outer container, not deep copy)
            # to conserve it.
            solver._index = index.copy()
        txn = solver.solve_for_transaction(prune=False, ignore_pinned=False)
        prefix_setup = txn.prefix_setups[prefix]
        return {
            "PREFIX": prefix,
            "LINK": [prec for prec in prefix_setup.link_precs],
        }


_install_actions = install_actions
del install_actions


@deprecated.argument("24.3", "24.5", "actions", rename="precs")
def _execute_actions(prefix, precs):
    # This is copied over from https://github.com/conda/conda/blob/23.11.0/conda/plan.py#L575
    # but reduced to only the functionality actually used within conda-build.
    assert prefix

    # Always link menuinst first/last on windows in case a subsequent
    # package tries to import it to create/remove a shortcut
    precs = [
        *(prec for prec in precs if prec.name == "menuinst"),
        *(prec for prec in precs if prec.name != "menuinst"),
    ]

    progressive_fetch_extract = ProgressiveFetchExtract(precs)
    progressive_fetch_extract.prepare()

    stp = PrefixSetup(prefix, (), precs, (), [], ())
    unlink_link_transaction = UnlinkLinkTransaction(stp)

    log.debug(" %s(%r)", "PROGRESSIVEFETCHEXTRACT", progressive_fetch_extract)
    progressive_fetch_extract.execute()
    log.debug(" %s(%r)", "UNLINKLINKTRANSACTION", unlink_link_transaction)
    unlink_link_transaction.execute()


@deprecated.argument("24.3", "24.5", "actions", rename="precs")
def _display_actions(prefix, precs):
    # This is copied over from https://github.com/conda/conda/blob/23.11.0/conda/plan.py#L58
    # but reduced to only the functionality actually used within conda-build.

    builder = ["", "## Package Plan ##\n"]
    if prefix:
        builder.append("  environment location: %s" % prefix)
        builder.append("")
    print("\n".join(builder))

    show_channel_urls = context.show_channel_urls

    def channel_str(rec):
        if rec.get("schannel"):
            return rec["schannel"]
        if rec.get("url"):
            return Channel(rec["url"]).canonical_name
        if rec.get("channel"):
            return Channel(rec["channel"]).canonical_name
        return UNKNOWN_CHANNEL

    def channel_filt(s):
        if show_channel_urls is False:
            return ""
        if show_channel_urls is None and s == DEFAULTS_CHANNEL_NAME:
            return ""
        return s

    packages = defaultdict(lambda: "")
    features = defaultdict(lambda: "")
    channels = defaultdict(lambda: "")

    for prec in precs:
        assert isinstance(prec, PackageRecord)
        pkg = prec["name"]
        channels[pkg] = channel_filt(channel_str(prec))
        packages[pkg] = prec["version"] + "-" + prec["build"]
        features[pkg] = ",".join(prec.get("features") or ())

    fmt = {}
    if packages:
        maxpkg = max(len(p) for p in packages) + 1
        maxver = max(len(p) for p in packages.values())
        maxfeatures = max(len(p) for p in features.values())
        maxchannels = max(len(p) for p in channels.values())
        for pkg in packages:
            # That's right. I'm using old-style string formatting to generate a
            # string with new-style string formatting.
            fmt[pkg] = f"{{pkg:<{maxpkg}}} {{vers:<{maxver}}}"
            if maxchannels:
                fmt[pkg] += " {channel:<%s}" % maxchannels
            if features[pkg]:
                fmt[pkg] += " [{features:<%s}]" % maxfeatures

    lead = " " * 4

    def format(s, pkg):
        return lead + s.format(
            pkg=pkg + ":",
            vers=packages[pkg],
            channel=channels[pkg],
            features=features[pkg],
        )

    if packages:
        print("\nThe following NEW packages will be INSTALLED:\n")
        for pkg in sorted(packages):
            print(format(fmt[pkg], pkg))
    print()
